import json
import logging
import logging.config
import logging.handlers
import os
import sys
import unicodedata

try:
    from StringIO import StringIO  ## for Python 2
except ImportError:
    from io import StringIO  ## for Python 3

from datetime import datetime
from decimal import Decimal, ROUND_HALF_UP
from threading import Lock
from time import sleep
from xml.etree import cElementTree as eTree

import tzlocal
from enum import Enum
from typing import Union, Dict


class IsoFormatter(logging.Formatter):
    def __init__(self, fmt=None, datefmt=None, style='%'):
        if fmt is None:
            fmt = "%(asctime)-6s: %(count)2s - %(name)s - %(levelname)s - "\
                  "%(thread)d:%(threadName)s - %(module)s:%(filename)s:%(lineno)s:%(funcName)s - "\
                  "%(message)s"
        super(IsoFormatter, self).__init__(fmt, datefmt)
        self.lock = Lock()
        self.count = 0
        self.max_count = 99

    def get_count(self):
        with self.lock:
            self.count += 1
            if self.count > self.max_count:
                self.count = 1
            return self.count

    def formatTime(self, record, datefmt=None):
        value = datetime.fromtimestamp(record.created, tzlocal.get_localzone())
        base_date = datetime.strftime(value, "%Y-%m-%dT%H:%M:%S")
        base_date += "." + str(int(record.msecs)).zfill(3)
        base_date += datetime.strftime(value, "%z")
        return base_date

    def format(self, record):
        if not hasattr(record, "count"):
            record.count = str(self.get_count()).zfill(2)
        return super(IsoFormatter, self).format(record)


def config_logger(loader_path,
                  log_name='component',
                  configure_root=False,
                  root_level=logging.WARN,
                  max_files=5,
                  max_size=5242880,
                  add_error_handler=False):
    if type(loader_path) == dict:
        logging.config.dictConfig(loader_path)
        return

    real_path = os.path.realpath(loader_path)
    if os.path.isfile(real_path) and real_path.endswith(".ini"):
        logging.config.fileConfig(real_path, disable_existing_loggers=False)
        return

    if os.path.isfile(real_path) and real_path.endswith(".json"):
        with open(real_path, "rb") as f:
            logging.config.dictConfig(json.loads(f.read().decode("utf-8")))
        return

    if os.path.isdir(real_path):
        dir_path = real_path
    else:
        dir_path = os.path.dirname(real_path)

    filename = os.path.join(dir_path, log_name + '.log')
    filename_error = os.path.join(dir_path, log_name + '.error.log')

    formatter = IsoFormatter()

    comp_logger = logging.getLogger(log_name)
    comp_logger.setLevel(logging.DEBUG)
    comp_logger.propagate = False

    rotating_file_logger = logging.handlers.RotatingFileHandler(filename, 'a', max_size, max_files)
    rotating_file_logger.setLevel(logging.DEBUG)
    rotating_file_logger.setFormatter(formatter)
    comp_logger.addHandler(rotating_file_logger)

    if add_error_handler:
        rotating_file_logger_error = logging.handlers.RotatingFileHandler(filename_error,
                                                                          'a',
                                                                          max_size,
                                                                          max_files)
        rotating_file_logger_error.setLevel(logging.ERROR)
        rotating_file_logger_error.setFormatter(formatter)
        comp_logger.addHandler(rotating_file_logger_error)

    if configure_root:
        root_logger = logging.getLogger()
        root_logger.addHandler(rotating_file_logger)
        root_logger.setLevel(root_level)


def import_pydevd(loader_path, port, suspend=False):
    try:
        current_dir = os.path.dirname(os.path.abspath(__file__))
        debug_lib_dir = os.path.join(current_dir, "..", "debug_lib")
        if os.path.exists(debug_lib_dir):
            sys.path.append(debug_lib_dir)
            from debughelper import import_pydevd as real_import_pydevd
            real_import_pydevd(loader_path, port, suspend)
    except ImportError:
        return


def get_path(directory='', file_name_with_extension=''):
    # type: (str, str) -> str
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), directory, file_name_with_extension)


def read_json_from_file_as_dictionary(json_file_path):
    with open(json_file_path, 'rb') as content_file:
        complete_json = content_file.read()
        return json.loads(complete_json.decode("utf-8"))


def remove_accents(text):
    # type: (Union[str, unicode]) -> str

    if not text:
        return text

    try:
        text = unicode(text, 'utf-8')
    except (TypeError, NameError):
        pass

    text = unicodedata.normalize('NFD', text)
    text = text.encode('ascii', 'ignore')
    return text


def json_serialize(obj, simple_enum=None):
    # type: (Any, bool) -> str
    try:
        return json.dumps(obj, default=lambda o: getattr(o, '__dict__', str(o) if not simple_enum else ""))
    except ValueError as _:
        return str(obj)


def json_pretty_format(dictionary):
    # type: (Dict) -> str

    return json.dumps(dictionary, sort_keys=True, indent=1, default=lambda o: getattr(o, '__dict__', str(o)))


def retry(logger, method_name, n, sleep_seconds, trying_to_do, is_success, success, failed, on_exception=None):
    for attempt in range(1, n + 1):
        # noinspection PyBroadException
        try:
            logger.debug("Retry [{}] :: trying_to_do {}/{}".format(method_name, attempt, n))
            response = trying_to_do()
            if is_success(response):
                logger.debug("Retry [{}] :: is_success {}/{}".format(method_name, attempt, n))
                return success(response)
            else:
                logger.debug("Retry [{}] :: handle_error {}/{}".format(method_name, attempt, n))
                failed(response)
        except Exception as _:
            logger.exception("Retry [{}] :: EXCEPTION {}/{}".format(method_name, attempt, n))
            if on_exception:
                on_exception()

        sleep(sleep_seconds)

    logger.error("Retry {} :: FAILED after {} tentatives".format(method_name, n))


def dict_to_object(dictionary, obj):
    # type: (dict, object) -> None

    for key in dictionary:
        if hasattr(obj, key):
            setattr(obj, key, dictionary[key])


def build_dict(seq, key):
    return dict((d[key], dict(d, index=index)) for (index, d) in enumerate(seq))


def get_class_by_name(kls):
    # type: (unicode) -> type
    parts = kls.split(".")
    module = ".".join(parts[:-1])
    m = __import__(module)
    for comp in parts[1:]:
        m = getattr(m, comp)
    return m


def remove_xml_namespace(xml):
    # type: (Union[str, unicode]) -> eTree.ElementTree
    it = eTree.iterparse(StringIO(xml))
    for _, el in it:
        if '}' in el.tag:
            el.tag = el.tag.split('}', 1)[1]
    return it.root


def round_half_away_from_zero(number, precision=0):
    number = Decimal(number)
    value = number.quantize(Decimal('10') ** -(precision + 2))
    return float(value.quantize(Decimal('10') ** -precision, rounding=ROUND_HALF_UP))


class ExtendedEnum(Enum):

    @classmethod
    def list_values(cls):
        # type: () -> list
        return list(map(lambda enum: enum.value, cls))

    @classmethod
    def list_names(cls):
        # type: () -> list
        return list(map(lambda enum: enum.name, cls))

    @classmethod
    def list_enums_tuple(cls):
        # type: () -> list
        return list(map(lambda enum: (enum.name, enum.value), cls))

    @classmethod
    def get_name(cls, value):
        # type: (int) -> Union[str, None]
        if value not in cls.list_values():
            return None

        return next(str(enum.name) for enum in cls if enum.value == value)

    @classmethod
    def get_value(cls, name):
        # type: (str) -> Union[int, None]
        if name not in cls.list_names():
            return None

        return next(int(enum.value) for enum in cls if enum.name == name)



from _OpenDaySummaryInfo import OpenDaySummaryInfo
from _OpenDaySummaryDto import OpenDaySummaryDto
from _OpenDaySummaryHeaderDto import OpenDaySummaryHeaderDto
from _OpenDaySummaryBodyDto import OpenDaySummaryBodyDto

__all__ = ["OpenDaySummaryInfo",
           "OpenDaySummaryDto",
           "OpenDaySummaryHeaderDto",
           "OpenDaySummaryBodyDto"]



# -*- coding: utf-8 -*-

import base64
import datetime
import json
import logging
import string

from xml.etree import cElementTree as eTree

import pyscripts
from actions.eft import finish_eft_transactions
from actions.events import event_publisher
from application.interactor import MessageDialogInteractor
from modelcommon.events import DialogEvent
from listeners import electronic_payment_listener
from pyscriptscache import cache as _cache
from sysactions import (
    get_model,
    translate_message,
    get_podtype,
    show_messagebox,
    get_posfunction,
    close_asynch_dialog,
    get_custom,
    show_any_dialog,
    update_any_dialog,
)

mbcontext = pyscripts.mbcontext
msrposmap = {}
finger_print_reader_dlg_id = -1
last_sitef_status_update = {}
logger = logging.getLogger("PosActions")


def _device_data_event_received(params):
    data, subject, p_type, asynch, pos_id = params[:5]
    try:
        logger.info("Data {0} ".format(data))

        xml = eTree.XML(data)
        device = xml.find("Device")
        device_name = str(device.get("name"))

        if device_name.startswith("scanner"):
            publish_scanner_event_to_ui(device, device_name)
        elif device_name.startswith("scale"):
            _get_weight(device_name, xml)
    except Exception as _:
        logger.exception('Exception processing device thread')


def publish_scanner_event_to_ui(device, device_name):
    pos_id = int(device_name[7:])
    barcode = base64.b64decode(device.text).strip('\0')

    event_subject = "POS{}".format(pos_id)
    event_publisher.publish_event(event_subject, "SCANNER_DATA", "SCANNER", barcode)


def _get_weight(device_name, xml):
    from posactions import pos_scale_events

    encoded_weight = xml.find("Device").text
    decoded_weight = base64.b64decode(encoded_weight)
    weight = float(decoded_weight) / 1000
    pos_scale_events.add_weight(device_name, weight)


def event_order_modified_sitef(params):
    xml, subject, event_type = params[:3]

    order = eTree.XML(xml).find(".//Order")
    pos_id = order.get("posId") or order.get("originatorId")[3:]
    model = get_model(pos_id)

    if event_type == "PAID":
        from posactions import is_table_service
        pod_type = get_podtype(model)
        pos_function = get_posfunction(model)
        is_table_service = is_table_service(model)
        if pod_type not in ("OT", "DL") and pos_function not in ("OT", "DL") and not is_table_service:
            from threading import Thread
            sitef_transactions_thread = Thread(target=finish_eft_transactions, args=(pos_id, order, "1"))
            sitef_transactions_thread.daemon = True
            sitef_transactions_thread.start()

    if event_type == "VOID_ORDER":
        from posactions import is_table_service
        pod_type = get_podtype(model)
        pos_function = get_posfunction(model)
        is_table_service = is_table_service(model)
        if pod_type not in ("OT", "DL") and pos_function not in ("OT", "DL") and not is_table_service:
            from threading import Thread
            sitef_transactions_thread = Thread(target=finish_eft_transactions, args=(pos_id, order, "0"))
            sitef_transactions_thread.daemon = True
            sitef_transactions_thread.start()


def add_or_update_default_options(query_template, queries, order_id, line_number, context, part_code, prod_qty, level, pod_type, ingredients=False, items=None, default_products=True, model=''):
    # Options - Pai tem default_qty = 1 e Filhos default_qty = 0
    # Ingredients - Pai tem default_qty = 0 e Filhos default_qty = 1
    all_options = [(_cache.get_tags_as_dict(part_code, "HasOptions"), 1, 0), (_cache.get_tags_as_dict(part_code, "Ingredients"), 0, 1)]
    forced_pricelist = get_custom(model, 'SERVICE_PERIOD') if model else False

    # Busca quais options devem estar apenas no OrderTaker
    dt_only_list = _cache.get_tags_as_dict(part_code, "DTOnly").split("|")
    dt_only_dict = {}
    for dt_only in dt_only_list:
        dt_only_dict[dt_only] = dt_only

    included_parts = {}
    if all_options:

        for options_line in all_options:
            if not options_line[0]:
                continue

            qty_tags = None
            temp_tags = _cache.get_tags_as_dict(part_code, "QtyOptions")
            if temp_tags:
                qty_tags = {tag.split(">")[0]: (tag.split(">")[1].split(";")[0], tag.split(">")[1].split(";")[1], tag.split(">")[1].split(";")[2]) for tag in temp_tags.split("|")}

            options = options_line[0].split("|")
            for opt in options:
                # OPTION - Code
                item_id = context + '.' + part_code
                temp_code = opt.split(">")[0]
                if pod_type != "DT" and temp_code in dt_only_dict:
                    continue

                included_parts[temp_code] = temp_code

                temp_level = level + 1
                inc_qty = qty_tags[temp_code][1] if qty_tags and temp_code in qty_tags else "1"
                ordered_qty = "NULL"
                if float(prod_qty) == 0:
                    default_qty = "0"
                else:
                    default_qty = qty_tags[temp_code][0] if qty_tags and temp_code in qty_tags else "1"
                price_key = _cache.get_best_price_key(item_id, temp_code, pod_type, forced_pricelist) or "NULL"
                queries.append(query_template % locals())

                if ingredients and options_line[2] == 1:
                    continue

                if default_products and len(opt.split(">")) > 1 and opt.split(">")[1]:
                    parent_item_id = item_id + '.' + temp_code
                    for product in opt.split(">")[1].split(";"):
                        # Default Product
                        item_id = parent_item_id
                        temp_code = product
                        temp_level = level + 2
                        inc_qty = prod_qty if float(prod_qty) == 0 else 1
                        ordered_qty = prod_qty if float(prod_qty) == 0 else 1
                        default_qty = options_line[2]
                        price_key = _cache.get_best_price_key(item_id, temp_code, pod_type, forced_pricelist) or "NULL"
                        queries.append(query_template % locals())

                        # Go to the NEXT LEVEL AND BEYOND
                        add_or_update_default_options(query_template, queries, order_id, line_number, item_id, temp_code, prod_qty, temp_level, pod_type, ingredients, items, model=model)

    parts = _cache.get_parts(part_code)
    if parts:
        for part in parts:
            if pod_type != "DT" and part in dt_only_dict:
                continue

            if part not in included_parts:
                item_id = context + '.' + part_code
                temp_code = part
                temp_level = level + 1
                inc_qty = prod_qty if float(prod_qty) == 0 else 1
                ordered_qty = prod_qty if float(prod_qty) == 0 else 1
                default_qty = float(parts[part][0] or prod_qty)
                price_key = _cache.get_best_price_key(item_id, temp_code, pod_type, forced_pricelist) or "NULL"

                # parts[part] -> Outros Elementos do Dicionário
                queries.append(query_template % locals())
                if items is not None:
                    price = _cache.get_best_price(item_id, temp_code, pod_type, forced_pricelist)
                    if price:
                        price = price[0]
                    items.append([item_id, temp_code, parts[part][3], price, line_number, prod_qty])

                # Go to the NEXT LEVEL
                add_or_update_default_options(query_template, queries, order_id, line_number, item_id, part, 1, temp_level, pod_type, ingredients, model=model)
# END of add_or_update_default_options


def update_options_and_defaults(queries, order_id, line_number, context, part_code, prod_qty, level, pod_type, model=''):
    temp_update = """UPDATE orderdb.CurrentOrderItem SET DefaultQty = %(default_qty)s
    WHERE OrderId = %(order_id)s AND LineNumber = %(line_number)s AND ItemId = '%(item_id)s' AND Level = %(temp_level)s AND PartCode = '%(temp_code)s'"""

    add_or_update_default_options(temp_update, queries, order_id, line_number, context, part_code, prod_qty, level, pod_type, False, model=model)
# END update_options_and_defaults


def insert_options_and_defaults(queries, order_id, line_number, context, part_code, prod_qty, level, pod_type, items=None, default_products=True, model=''):
    temp_insert = """INSERT OR REPLACE INTO orderdb.CurrentOrderItem (OrderId, LineNumber, ItemId, Level, PartCode, OrderedQty, IncQty, DefaultQty, PriceKey)
    VALUES (%(order_id)s, %(line_number)s, '%(item_id)s', %(temp_level)s, '%(temp_code)s', %(ordered_qty)s, %(inc_qty)s, %(default_qty)s, %(price_key)s)"""

    add_or_update_default_options(temp_insert, queries, order_id, line_number, context, part_code, prod_qty, level, pod_type, True, items, default_products, model=model)
# END of insert_options_and_defaults


def delete_options_and_defaults(queries, order_id, line_number, context, part_code, prod_qty, level, pod_type, items=None, model=''):
    temp_delete = """DELETE FROM orderdb.CurrentOrderItem
    WHERE OrderId = %(order_id)s AND LineNumber = %(line_number)s AND ItemId LIKE '%(item_id)s%%' AND Level >= %(temp_level)s"""

    add_or_update_default_options(temp_delete, queries, order_id, line_number, context, part_code, prod_qty, level, pod_type, False, items, model=model)
# END of delete_options_and_defaults


def get_updated_sale_line_defaults(queries, order_id, line_number, context, part_code, prod_qty, level, pod_type, model=''):
    temp_lines = """
        {
            "order_id": "%(order_id)s",
            "line_number": "%(line_number)s",
            "item_id": "%(item_id)s",
            "level": "%(temp_level)s",
            "part_code": "%(temp_code)s",
            "ordered_qty": "%(ordered_qty)s",
            "inc_qty": "%(inc_qty)s",
            "default_qty": "%(default_qty)s",
            "price_key": "%(price_key)s"
        }
    """
    add_or_update_default_options(temp_lines, queries, order_id, line_number, context, part_code, prod_qty, level, pod_type, False, model=model)
# END of get_updated_sale_line_defaults


def event_pre_new_order(params):
    queries = ["""DELETE FROM orderdb.CurrentOrderItem"""]

    if queries:
        return "\0".join(["0"] + [";".join(queries)])
    return None


def eft_status_update(params):
    message, pos_id = params[0].split(';')
    pos_id = str(int(pos_id))

    dialog_type = "messageDialog"
    title = "$INFORMATION"
    timeout = 60000

    if pos_id not in last_sitef_status_update or last_sitef_status_update[pos_id] is None:
        dialog_id = show_any_dialog(pos_id, dialog_type, message, title, "", "", "", timeout, "NOCONFIRM", "", "", "", None, True)
        last_sitef_status_update[pos_id] = dialog_id
    else:
        dialog_id = last_sitef_status_update[pos_id]
        last_sitef_status_update[pos_id] = dialog_id
        update_any_dialog(pos_id, dialog_id, dialog_type, message, title, "", "", "", timeout, "NOCONFIRM", "", "", "", None)


def eft_status_finish(params):
    pos_id = params[0].split(';')[0]
    pos_id = str(int(pos_id))

    if pos_id in last_sitef_status_update and last_sitef_status_update[pos_id] is not None:
        close_asynch_dialog(pos_id, last_sitef_status_update[pos_id])
        last_sitef_status_update[pos_id] = None


def wait_message_dialog(params):
    event_type = params[2]
    data = json.loads(params[0])
    MessageDialogInteractor(data, event_type).wait_messages()


def finger_print_reader_callback(params):
    global finger_print_reader_dlg_id

    params_list = params[0].split(";")
    pos_id = params_list[0]
    extra_data = None
    if len(params_list) > 1:
        extra_data = params_list[1]

    event_type = params[2]

    message = None
    title = "$FINGERPRINT"
    timeout = 300000

    if event_type == "EVT_CLOSE_ALL_DIALOGS":
        pass
    elif event_type == "EVT_FINGER_ALREADY_REGISTERED":
        title = "$ERROR"
        message = "$FINGERPRINT_ALREADY_FOR_USER"
        timeout = 3000
    elif event_type == "EVT_BAD_FINGERPRINT_READ":
        title = "$ERROR"
        message = "$INVALID_FINGERPRINT_READING"
    elif event_type == "EVT_PLACE_RIGHT_INDEX_FINGER":
        message = "$PLACE_RIGHT_INDEX_FINGER"
    elif event_type == "EVT_PLACE_FINGER_LOGIN":
        title = "$PLACE_FINGER_LOGIN"
        message = "$PLACE_FINGER_LOGIN"
    elif event_type == "EVT_PLACE_FINGER_AUTHORIZATION":
        title = "$TITLE_PLACE_FINGER_AUTHORIZATION"
        message = "$PLACE_FINGER_AUTHORIZATION"
    elif event_type == "EVT_FINGER_READ_SUCCESSFULLY":
        message = "$FINGER_READ_SUCCESSFULLY"
    elif event_type == "EVT_FINGER_READ_SUCCESSFULLY_READ_AGAIN":
        if extra_data == '1':
            message = translate_message(get_model(str(pos_id)), "FIRST_READING")
        elif extra_data == '2':
            message = translate_message(get_model(str(pos_id)), "SECOND_READING")
        elif extra_data == '3':
            message = translate_message(get_model(str(pos_id)), "THIRD_READING")
        elif extra_data == '4':
            message = translate_message(get_model(str(pos_id)), "FOURTH_READING")
        elif extra_data == '5':
            message = translate_message(get_model(str(pos_id)), "FIFTH_READING")
        elif extra_data == '6':
            message = translate_message(get_model(str(pos_id)), "SIXTH_READING")
        elif extra_data == '7':
            message = translate_message(get_model(str(pos_id)), "SEVENTH_READING")
        elif extra_data == '8':
            message = translate_message(get_model(str(pos_id)), "EIGHTH_READING")
        elif extra_data == '9':
            message = translate_message(get_model(str(pos_id)), "NINTH_READING")
        elif extra_data == '10':
            message = translate_message(get_model(str(pos_id)), "TENTH_READING")
        elif extra_data == '11':
            message = translate_message(get_model(str(pos_id)), "ELEVENTH_READING")
        else:
            message = ""
        message = "{} {}".format(message, translate_message(get_model(str(pos_id)), "SUCCESS_PLACE_YOUR_FINGER_AGAIN"))
    elif event_type == "EVT_ENROLLMENT_SUCCESS":
        message = "$REGISTRATION_PERFORMED_SUCCESSFULLY"
        timeout = 3000
    else:
        message = "$EVENT_NOT_REGISTERED"

    new_id = None
    if message is not None:
        logger.info("Fingerprint callback message: {}".format(message))
        new_id = show_messagebox(pos_id, message, title, "info", buttons="", timeout=timeout, asynch=True)

    if finger_print_reader_dlg_id != -1:
        close_asynch_dialog(pos_id, finger_print_reader_dlg_id)

    if new_id is not None:
        finger_print_reader_dlg_id = new_id
# END def finger_print_reader_callback


def get_whooper_wifi_code(bk_number, order_id, date):
    alphabet = string.digits + string.ascii_uppercase

    cc = str(bk_number).zfill(5)[-5:] + str(order_id).zfill(4)[-4:] + datetime.datetime.strftime(date, "%d%m%y")

    scrambled_concat = cc[9] + cc[8] + cc[1] + cc[7] + cc[11] + cc[14] + cc[2] + cc[3] + cc[5] + cc[6] + cc[4] + cc[10] + cc[12] + cc[13] + cc[0]
    simplified_scrambled_concat = cc[9] + cc[10] + cc[8] + cc[1] + cc[7] + cc[2] + cc[3] + cc[5] + cc[6] + cc[4] + cc[0]

    bases10 = []
    for i in range(0, 11):
        bases10.append(int(simplified_scrambled_concat[i]) * pow(10, 10 - i))

    the_number = sum(bases10)

    code_number = []
    for i in range(6, -1, -1):
        code_number.append((the_number / pow(36, i)) % 36)

    multiple = 15
    divisor_number_sum = 0
    while multiple > 0:
        for number in scrambled_concat:
            divisor_number_sum += int(number) * multiple
            multiple -= 1
    code_number.append(divisor_number_sum % 36)

    code_alphanumeric = []
    for i in code_number:
        code_alphanumeric.append(alphabet[i])

    final_code = ''.join(code_alphanumeric)
    return final_code


#
# Main function (called by pyscripts)
#
def main():
    pyscripts.subscribe_event_listener("ORDER_MODIFIED", event_order_modified_sitef)
    pyscripts.subscribe_event_listener("PRE_NEW_ORDER", event_pre_new_order)
    pyscripts.subscribe_event_listener("SITEF_STATUS_UPDATE", eft_status_update)
    pyscripts.subscribe_event_listener("SITEF_STATUS_UPDATE_FINISHED", eft_status_finish)
    pyscripts.subscribe_event_listener("FPR_CALLBACK", finger_print_reader_callback)
    pyscripts.subscribe_event_listener("DEVICE_DATA", _device_data_event_received)
    electronic_payment_listener.init()
    pyscripts.subscribe_event_listener(DialogEvent.WAIT_MESSAGE_DIALOG, wait_message_dialog)



class InvalidOrderEventException(RuntimeError):
    def __init__(self, data):
        message = 'Invalid Order Event. Event: "{}"'.format(data)
        super(InvalidOrderEventException, self).__init__(message)



import logging
import time
import json
from threading import Thread

from bustoken import TK_POS_UPDATER_GET_VERSIONS
from msgbus import TK_SYS_ACK
from sysactions import action, send_message

last_pos_version = ""
pos_version_thread = None

logger = logging.getLogger("PosActions")


@action
def get_pos_versions(_):
    global pos_version_thread

    if pos_version_thread is None:
        pos_version_thread = Thread(target=_get_pos_version)
        pos_version_thread.daemon = True
        pos_version_thread.start()

    return last_pos_version


def _get_pos_version():
    global last_pos_version

    while True:
        try:
            reply = send_message("PosUpdater", TK_POS_UPDATER_GET_VERSIONS, timeout=5000000)
            if reply.token == TK_SYS_ACK:
                last_pos_version = reply.data
        except Exception as _:
            logger.exception("Error getting POS Versions")
        finally:
            time.sleep(5)



from typing import List  # noqa
from _PickListItem import PickListItem  # noqa


class PickListBodyDto(object):
    def __init__(self, pick_list_items):
        # type: (List[PickListItem]) -> ()
        self.pick_list_items = pick_list_items



# -*- coding: utf-8 -*-
import time
import json

from xml.etree import cElementTree as eTree
from typing import Optional, List

from sysactions import show_messagebox, StopAction, get_model, get_current_order, get_posot

from actions.coupon.display_coupon import get_benefits_from_order
from actions.eft import finish_eft_transactions
from application.interactor import (
    NDiscountInteractor,
    OrderInteractor,
)


def clean_discounts(pos_id):
    model = get_model(pos_id)
    order = model.find(".//Order")
    clean_orders_discounts(pos_id, [order])

    return order


def clean_discounts_and_payments(pos_id, disconnected_discount_component=False):
    model = get_model(pos_id)
    order = get_current_order(model)

    clean_orders_discounts(pos_id, [order], disconnected_discount_component)
    return clean_order_payments(
        pos_id=pos_id,
        order=order,
    )


def has_manual_discounts(
        order   # type: eTree.XML
):
    # type: (...) -> bool

    if not order:
        return False

    custom_properties = order.find(".//CustomOrderProperties")
    benefit_list_custom_property = custom_properties.find(".//OrderProperty[@key='BENEFIT_LIST']")
    if benefit_list_custom_property is None:
        return False

    benefits = json.loads(benefit_list_custom_property.get("value"))
    return True if benefits["manualBenefitOnTotalSequence"] else False


def clean_order_payments(
        pos_id,                         # type: int
        order,                          # type: eTree.XML
        check_manual_discounts=False,   # type: Optional[bool]
        store=False                     # type: Optional[bool]
):
    # type: (...) -> Optional[bool]

    has_discounts = check_manual_discounts and has_manual_discounts(order)
    order_tenders = filter(
        lambda t: float(t.get("tenderAmount")) > 0,
        order.findall(".//TenderHistory/Tender")
    )
    order_id = order.get('orderId')
    order_interactor = OrderInteractor(pos_id=str(pos_id))
    if len(order_tenders) >= 1:
        if has_discounts:
            message = "$CLEAN_TENDERS_AND_MANUAL_DISCOUNTS_TO_EXECUTE_ACTION"
            message = "$CLEAN_TENDERS_AND_MANUAL_DISCOUNTS_TO_STORE_ORDER" if store else message
            ret = show_messagebox(pos_id, message, buttons="$YES|$NO")
        else:
            message = "$CLEAN_TENDERS_TO_STORE_ORDER" if store else "$CLEAN_TENDERS_TO_EXECUTE_ACTION"
            ret = show_messagebox(pos_id, message, buttons="$YES|$NO")
        if ret != 0:
            raise StopAction()

        finish_eft_transactions(pos_id, order, "0")
        for tender in order_tenders:
            order_interactor.clear_order_tender(
                order_tender_id=tender.get("tenderId"),
                order_id=order_id,
            )
        return True
    if has_discounts:
        message = "$CLEAN_MANUAL_DISCOUNTS_TO_STORE_ORDER" if store else "$CLEAN_MANUAL_DISCOUNTS_TO_EXECUTE_ACTION"
        ret = show_messagebox(pos_id, message, buttons="$YES|$NO")

        if ret != 0:
            raise StopAction()

    return True


def clean_orders_discounts(
    pos_id,                                     # type: str
    orders,                                     # type: eTree.XML
    disconnected_discount_component=False,      # type: Optional[bool]
    message="$CONFIRM_CLEAN_DISCOUNT"           # type: Optional[str]
):
    # type: (...) -> Optional[bool, object]

    model = get_model(pos_id)
    pos_ot = get_posot(model)

    discount_ids = []
    lines_to_void = {}

    for order in orders:
        lines_to_void = _get_lines_with_coupons(order, lines_to_void)
        discount_ids.extend(_get_lines_discount_ids(order))

    if len(lines_to_void) == 0 and not discount_ids:
        return

    ret = show_messagebox(
        pos_id,
        message,
        buttons="$YES|$NO"
    )
    if ret != 0:
        raise StopAction()

    if disconnected_discount_component:
        success = _clean_static_orders_discounts(pos_id, pos_ot, discount_ids, lines_to_void)
        time.sleep(0.5)
        return success
    elif len(discount_ids) > 0:
        success = _clean_discount_with_id(pos_ot, discount_ids)
        return success
    else:
        return NDiscountInteractor.remove_benefits(pos_id)


def _clean_static_orders_discounts(pos_id, pos_ot, discount_ids, lines_to_void):
    # type: (str, object, List, dict) -> bool

    try:
        if len(lines_to_void) > 0:
            pos_ot.blkopnotify = True
            for order_id in lines_to_void:
                pos_ot.voidLine(int(pos_id), "|".join(lines_to_void[order_id]))
            pos_ot.blkopnotify = False

        _clean_discount_with_id(pos_ot, discount_ids)
        return True
    except Exception as e:
        raise e.message


def _clean_discount_with_id(pos_ot, discount_ids):
    # type: (object, List) -> bool

    try:
        for discount in list(dict.fromkeys(discount_ids)):
            pos_ot.clearDiscount(discount)
        return True
    except Exception as e:
        raise e.message


def _get_lines_with_coupons(order, lines_to_void=None):
    if lines_to_void is None:
        lines_to_void = {}

    sale_lines = order.findall(".//SaleLine")
    for sale_line in sale_lines:
        item_type = sale_line.get("itemType")
        qty = float(sale_line.get('qty'))
        if item_type == "COUPON" and qty > 0:
            line_number = sale_line.get("lineNumber")
            order_id = int(order.get("orderId"))
            if order_id in lines_to_void:
                lines_to_void[order_id].append(line_number)
            else:
                lines_to_void[order_id] = [line_number]
    return lines_to_void


def _get_lines_discount_ids(order):
    automatic_discount_ids = _get_automatic_discount_ids(order)
    discounts = []
    sale_lines = order.findall(".//SaleLine")
    for sale_line in sale_lines:
        # discountsApplied is like this in teh CORE: "2:0.9,1:5.00"
        discounts_applied = sale_line.get("discountsApplied")
        if discounts_applied:
            for discount in discounts_applied.split(','):
                discount_id = discount.split(':')[0]
                if discount_id != '2' and int(discount_id) not in automatic_discount_ids:
                    discounts.append(discount_id)

    return discounts


def _get_automatic_discount_ids(order):
    automatic_discount_ids = []
    order_custom_properties = order.find(".//CustomOrderProperties")
    if order_custom_properties:
        benefits = get_benefits_from_order(None, order, order_custom_properties)
        for benefit in benefits:
            automatic_benefit = benefit.get("benefit") and benefit.get("benefit").get("automatic") != 'DISABLED'
            used_discount_ids = benefit.get("used_discount_ids")
            if not automatic_benefit or not used_discount_ids:
                continue

            automatic_discount_ids.extend(benefit["used_discount_ids"])

    return automatic_discount_ids



# -*- coding: utf-8 -*-
from externals.util import encode_utf_8


class Address(object):
    def __init__(self, address_id, zip_code, city, street, neighborhood, number, complement, reference_point):
        # type: (str, str, str, str, str, str, str, str) -> None

        self.address_id = address_id
        self.zip_code = zip_code
        self.city = encode_utf_8(city)
        self.street = encode_utf_8(street)
        self.neighborhood = encode_utf_8(neighborhood)
        self.number = encode_utf_8(number)
        self.complement = encode_utf_8(complement)
        self.reference_point = encode_utf_8(reference_point)



from datetime import date

from actions.update import get_catalog_version_to_apply, get_message_update, apply_update
from actions.rupture.get_ruptured_products import has_rupture_items
from actions.rupture.clean_rupture import clean_rupture_on_open_day
from msgbus import TK_SYS_ACK
from sysactions import (
    show_messagebox,
    format_date,
    StopAction,
    get_model,
    get_poslist,
    get_business_period,
)

BUTTONS = "$OK|$CANCEL"


class OpenDayInteractor(object):
    def process_before_open_day(
        self,
        pos_id,  # type: int
    ):
        # type: (...) -> None

        self._check_has_update_menu(pos_id)

        if self._is_first_pos_opening_current_day():
            clean_rupture_on_open_day()

    @staticmethod
    def _check_has_update_menu(
        pos_id,  # type: int
    ):
        # type: (...) -> None

        today = date.today().strftime("%Y%m%d")
        confirm_message = "$CONFIRM_OPEN_CURRENT_DAY|{0}".format(format_date(today))
        msg = get_catalog_version_to_apply()
        if msg.token == TK_SYS_ACK:
            get_message_update(pos_id, "$APPLYING_UPDATE_MENU_VERSION_PENDING|" + msg.data)
            has_rupture_items(pos_id)
            applied_update = apply_update(pos_id)
            if not applied_update:
                confirm_message = "$UPDATE_MENU_ERROR_PROCEED"

        confirm = show_messagebox(pos_id, message=confirm_message, buttons=BUTTONS)
        if (confirm is None) or (confirm == 1):
            raise StopAction()

    @staticmethod
    def _is_first_pos_opening_current_day():
        # type: (...) -> bool

        pos_list = sorted(get_poslist())
        current_business_period = date.today().strftime("%Y%m%d")

        for pos in pos_list:
            model = get_model(posid=pos)
            business_period = get_business_period(model=model)
            if current_business_period == business_period:
                return False

        return True



import sys
from msgbus import MBEasyContext  # noqa
from xml.etree import cElementTree as etree
from xml.etree.ElementTree import Element  # noqa
from domain import Clock  # noqa

from commons.report import Generator
from reports_app.picklist.dto import PickListBodyDto
from reports_app.picklist.dto import PickListDto
from reports_app.picklist.dto import PickListHeaderDto
from reports_app.picklist.dto import PickListItem


class PickListGenerator(Generator):
    def __init__(self, mb_context, pick_list_xml, clock, store_number):
        # type: (MBEasyContext, Clock, unicode, unicode) -> None
        super(PickListGenerator, self).__init__()
        self.mb_context = mb_context
        self.clock = clock
        self.store_number = store_number
        self.pick_list_xml = pick_list_xml

    def generate_data(self):
        # type: () -> PickListDto
        order, pod_type, sale_type, pos_id = self._validate_xml_and_get_data()
        pick_list_header = self._generate_header(order)
        pick_list_body = PickListBodyDto(self._generate_body(order))
        return PickListDto(pick_list_header, pick_list_body, pod_type, sale_type, pos_id)

    def _validate_xml_and_get_data(self):
        order = self._validate_xml_and_get_order()
        pod_type = self._validate_pod_type(order)
        pos_id = self._validate_pos_id(order)
        sale_type = self._validate_sale_type(order)
        return order, pod_type, sale_type, pos_id

    def _validate_xml_and_get_order(self):
        try:
            xml = etree.XML(self.pick_list_xml)
            order = xml
        except Exception as ex:
            raise PickListGenerator.InvalidXml(ex.message), None, sys.exc_info()[2]
        return order

    def _validate_pod_type(self, order):
        if "pod_type" not in order.attrib:
            raise PickListGenerator.InvalidXml(PickListGenerator.InvalidXml.NoOrderPodTypeMessage)
        pod_type = order.attrib["pod_type"]
        return pod_type

    def _validate_pos_id(self, order):
        if "pos_id" not in order.attrib:
            raise PickListGenerator.InvalidXml(PickListGenerator.InvalidXml.NoOrderPosIdMessage)
        pos_id = order.attrib["pos_id"]
        return pos_id

    def _validate_sale_type(self, order):
        if "sale_type" not in order.attrib:
            raise PickListGenerator.InvalidXml(PickListGenerator.InvalidXml.NoOrderSaleTypeMessage)
        sale_type = order.attrib["sale_type"]
        return sale_type

    def _generate_header(self, order):
        if "order_id" not in order.attrib:
            raise PickListGenerator.InvalidXml(PickListGenerator.InvalidXml.NoOrderIdMessage)
        store = self.store_number
        order_id = order.attrib["order_id"]
        name = self._get_customer_name(order)

        return PickListHeaderDto(self.clock.get_current_time(), order_id, store, name)

    @staticmethod
    def _get_customer_name(order):
        name = None
        name_node = order.find("Properties/Property[@key='CUSTOMER_NAME']")
        if name_node is not None:
            name = name_node.attrib["value"]
        return name

    @staticmethod
    def _generate_body(order):
        # type: (Element) -> PickListBodyDto
        items_node = order.find("Items")
        if items_node is None:
            raise PickListGenerator.InvalidXml(PickListGenerator.InvalidXml.OrderWithoutItems)

        items = items_node.findall("Item")
        if len(items) == 0:
            raise PickListGenerator.InvalidXml(PickListGenerator.InvalidXml.OrderWithoutItems)
        pick_list_items = []

        for item in items:
            pick_list_item = PickListItem(item)
            if pick_list_item.description is None:
                raise PickListGenerator.InvalidXml(PickListGenerator.InvalidXml.ItemWithoutDescription)
            if pick_list_item.qty is None:
                raise PickListGenerator.InvalidXml(PickListGenerator.InvalidXml.ItemWithoutQuantity)
            if pick_list_item.default_qty is None:
                raise PickListGenerator.InvalidXml(PickListGenerator.InvalidXml.ItemWithoutDefaultQuantity)
            if pick_list_item.item_type is None:
                raise PickListGenerator.InvalidXml(PickListGenerator.InvalidXml.ItemWithoutItemType)

            pick_list_items.append(pick_list_item)

        return pick_list_items

    class InvalidXml(Exception):
        NoOrderIdMessage = "Order without order_id attribute"
        NoOrderPodTypeMessage = "Order without pod_type attribute"
        NoOrderPosIdMessage = "Order without pos_id attribute"
        OrderWithoutItems = "Order does not have items"
        ItemWithoutDescription = "Item without description: "
        ItemWithoutQuantity = "Item without quantity: "
        ItemWithoutDefaultQuantity = "Item without default quantity: "
        ItemWithoutItemType = "Item without item type: "
        NoOrderSaleTypeMessage = "Order without sale_type attribute"

        def __init__(self, message):
            super(PickListGenerator.InvalidXml, self).__init__(message)



from typing import Dict


class ApiRequest(object):
    def __init__(self, method, api_path, path_parameters, query_parameters, headers, body):
        # type: (str, str, Dict[str, str], Dict[str, str], Dict[str, str], bytes) -> None
        self.method = method
        self.api_path = api_path
        self.path_parameters = path_parameters
        self.query_parameters = query_parameters
        self.headers = headers
        self.body = body

    def __str__(self):
        body = self.body
        if isinstance(self.body, str):
            body = self.body.decode("utf-8")
        return u"ApiRequest('{}', {}, {}, '{}', {}, '{}')".format(self.method,
                                                                  self.api_path,
                                                                  self.path_parameters,
                                                                  self.query_parameters,
                                                                  self.headers,
                                                                  body)

    def __repr__(self):
        return self.__str__()



import re
import urllib2
import random
import dns.resolver

from application.model.configuration import Configurations


class LoyaltyAvailableInteractor(object):
    def __init__(self, configs):
        # type: (Configurations) -> None

        self.logger = configs.logger
        self.url_to_test = configs.base_url

    def is_loyalty_available(self):
        # type: () -> bool
        try:
            self.has_dns_access()
            self.logger.debug('Url to test: "{}"'.format(self.url_to_test))
            urllib2.urlopen(self.url_to_test, timeout=20)
            self.logger.debug('Test OK')
            return True
        except urllib2.HTTPError as e:
            self.logger.debug('Test OK, code: {}'.format(e.code))
            return True
        except urllib2.URLError as e:
            self.logger.debug('Test failed with reason: "{}"'.format(e.reason))
            return False
        except ValueError as e:
            self.logger.debug('DNS Server is unreacheble: {}'.format(e))

    def has_dns_access(self):
        name_servers = dns.resolver.Resolver().nameservers
        timeout = 5
        url_without_protocol = re.sub('^https?://', '', self.url_to_test)

        if name_servers:
            dns_server = random.choice(name_servers)
            self.logger.debug('Test access DNS Server: {}'.format(dns_server))
            request = dns.message.make_query(url_without_protocol, dns.rdatatype.A)
            dns_query = dns.query.tcp(request, dns_server, timeout).rcode()
            if dns_query == 0:
                return True
            else:
                raise ValueError('DNS Error')

        self.logger.debug('DNS Servers not configured')
        return False



# encoding: utf-8
import os
import iso8601
from helper import MwOrderStatus
from old_helper import BaseRepository
from msgbus import MBEasyContext
from ..model import Order, OrderTender
from persistence import Connection
from datetime import datetime, timedelta
from typing import List, Dict, Optional
from xml.etree import cElementTree as eTree
from systools import sys_log_exception


class OrderRepository(BaseRepository):
    TypeRealDate = "RealDate"
    TypeBusinessPeriod = "BusinessPeriod"
    TypeSessionId = "SessionId"

    def __init__(self, mbcontext, pos_list, fiscal_sent_dir):
        # type: (MBEasyContext, List[int], unicode) -> None
        super(OrderRepository, self).__init__(mbcontext)
        self.pos_list = pos_list
        self.fiscal_sent_dir = fiscal_sent_dir

    def get_paid_orders_by_real_date(self, initial_date, end_date, operator_id, report_pos, session_id=None):
        # type: (datetime, datetime, Optional[unicode], Optional[unicode]) -> List[Order]
        return self._get_paid_orders(self.TypeRealDate, initial_date, end_date, operator_id, session_id, report_pos)

    def get_paid_orders_by_business_period(self, initial_date, end_date, operator_id, report_pos):
        # type: (datetime, datetime, unicode, unicode) -> List[Order]
        return self._get_paid_orders(self.TypeBusinessPeriod, initial_date, end_date, operator_id, None, report_pos)

    def get_voided_orders_by_real_date(self, initial_date, end_date, operator_id, report_pos, session_id=None):
        # type: (datetime, datetime, unicode, unicode) -> List[Order]
        return self._get_voided_orders(self.TypeRealDate, initial_date, end_date, operator_id, session_id, report_pos)

    def get_voided_orders_by_business_period(self, initial_date, end_date, operator_id, report_pos):
        # type: (datetime, datetime, unicode, unicode) -> List[Order]
        return self._get_voided_orders(self.TypeBusinessPeriod, initial_date, end_date, operator_id, None, report_pos)

    def get_paid_orders_by_session_id(self, session_id):
        # type: (unicode) -> List[Order]
        return self._get_paid_orders(self.TypeSessionId, None, None, None, session_id, None)

    def get_voided_orders_by_session_id(self, session_id):
        # type: (unicode) -> List[Order]
        return self._get_voided_orders(self.TypeSessionId, None, None, None, session_id, None)

    def get_paid_orders_by_xml(self, initial_date, end_date):
        # type: (datetime, datetime) -> List[Order]
        orders = []  # type: List[Order]

        current_date = initial_date
        while current_date <= end_date:
            xml_directory = os.path.join(self.fiscal_sent_dir, "Enviados", current_date.strftime("%Y"), current_date.strftime("%m"), current_date.strftime("%d"))
            if os.path.exists(xml_directory):
                for entry in os.listdir(xml_directory):
                    try:
                        if entry.endswith(".xml"):
                            with open(os.path.join(xml_directory, entry), "rb") as xml_file:
                                xml_content = xml_file.read()
                            order = self._get_order_from_xml(entry, xml_content)
                            if order is not None:
                                orders.append(order)
                    except Exception as ex:
                        sys_log_exception("Fail to read xml: %s; Exception: %s" % (xml_directory, ex))

            current_date += timedelta(days=1)

        return orders

    def get_voided_orders_by_xml(self, initial_date, end_date):
        # type: (datetime, datetime) -> List[Order]
        return []

    def get_hourly_sale(self, report_pos, initial_date, end_date, operator):
        # type: (unicode, Optional[datetime], Optional[datetime]) -> List[object]
        def inner_func(conn):
            # type: (Connection) -> List[object]
            try:
                initial_date_formatted = initial_date[:4] + '-' + initial_date[4:6] + '-' + initial_date[6:]
                end_date_formatted = end_date[:4] + '-' + end_date[4:6] + '-' + end_date[6:]
                query = self._QueryHourlySale.format(initial_date_formatted, end_date_formatted, operator)

                order_by_hour = [(str(x.get_entry(0)),
                                 str(x.get_entry(1)),
                                 int(x.get_entry(2)),
                                 int(x.get_entry(3)),
                                 float(x.get_entry(4)),
                                 int(x.get_entry(5))) for x in conn.select(query)]

                return list(order_by_hour)
            except Exception as ex:
                sys_log_exception('erro: %s' % ex)
                return list()

        report_pos_list = self.pos_list if report_pos is None else (report_pos,)
        return self.execute_parallel_in_all_databases_returning_flat_list(inner_func, report_pos_list)

    def _get_paid_orders(self, report_type, initial_date, end_date, operator_id, session_id, report_pos):
        # type: (unicode, Optional[datetime], Optional[datetime], Optional[unicode], Optional[unicode]) -> List[Order]

        def inner_func(conn):
            # type: (Connection) -> List[Order]

            try:
                if operator_id is not None:
                    current_session_id = "user={},".format(operator_id)
                else:
                    current_session_id = session_id if session_id is not None and session_id != 'None' else ''

                if report_type == self.TypeRealDate:
                    query = self._PaidOrdersByRealDateQuery
                    date_format = "%Y-%m-%d"
                    formatted_initial_date = initial_date.strftime(date_format)
                    formatted_end_date = end_date.strftime(date_format)
                    query = query.format(formatted_initial_date, formatted_end_date, current_session_id)
                elif report_type == self.TypeBusinessPeriod:
                    query = self._PaidOrderByBusinessPeriod
                    date_format = "%Y%m%d"
                    formatted_initial_date = initial_date.strftime(date_format)
                    formatted_end_date = end_date.strftime(date_format)
                    query = query.format(formatted_initial_date, formatted_end_date, current_session_id)
                else:
                    query = self._PaidOrderBySessionId
                    date_format = "%Y%m%d"
                    query = query.format(session_id)

                all_orders = {}  # type: Dict[int, Order]
                order_with_payments = [(int(x.get_entry(0)),
                                        int(x.get_entry(1)),
                                        datetime.strptime(x.get_entry(2), date_format),
                                        float(x.get_entry(3)),
                                        int(x.get_entry(4)),
                                        float(x.get_entry(5)),
                                        float(x.get_entry(6)),
                                        x.get_entry(7),
                                        x.get_entry(8),
                                        x.get_entry(9),
                                        x.get_entry(10),
                                        x.get_entry(11)) for x in conn.select(query)]

                for order_tuple in order_with_payments:
                    order_id = order_tuple[0]
                    order_state = order_tuple[1]
                    order_date = order_tuple[2]
                    order_total = order_tuple[3]
                    donation_value = order_tuple[6]
                    discount_value = float(order_tuple[7]) if order_tuple[7] else None
                    tip_value = float(order_tuple[8]) if order_tuple[8] else None
                    sale_type = int(order_tuple[9])
                    table_id = order_tuple[10]
                    delivery_originator = order_tuple[11]

                    if order_id in all_orders:
                        order = all_orders[order_id]
                    else:
                        order = Order(order_id, order_date, order_state, order_total, donation_value, [], None, discount_value, tip_value, sale_type, table_id, None, delivery_originator)
                        all_orders[order_id] = order

                    tender_type = order_tuple[4]
                    tender_value = order_tuple[5]

                    order_tender = OrderTender(tender_type, tender_value)
                    order.tenders.append(order_tender)
                return list(all_orders.values())
            except Exception as _:
                return list({"error": "error"}.values())

        report_pos_list = self.pos_list if report_pos is None else (report_pos,)
        return self.execute_parallel_in_all_databases_returning_flat_list(inner_func, report_pos_list)

    def _get_voided_orders(self, report_type, initial_date, end_date, operator_id, session_id, report_pos):
        # type: (unicode, Optional[datetime], Optional[datetime], Optional[unicode], Optional[unicode], Optional[unicode]) -> List[Order]
        def inner_func(conn):
            # type: (Connection) -> List[Order]
            try:
                if operator_id is not None:
                    current_session_id = "user={},".format(operator_id)
                else:
                    current_session_id = session_id if session_id is not None and session_id != 'None' else ''

                if report_type == self.TypeRealDate:
                    query = self._VoidedOrdersByRealDateQuery
                    date_format = "%Y-%m-%d"
                    query = query.format(initial_date.strftime(date_format), end_date.strftime(date_format), current_session_id)
                elif report_type == self.TypeBusinessPeriod:
                    query = self._VoidedOrdersByBusinessPeriod
                    date_format = "%Y%m%d"
                    query = query.format(initial_date.strftime(date_format), end_date.strftime(date_format), current_session_id)
                else:
                    query = self._VoidedOrdersBySessionId
                    date_format = "%Y%m%d"
                    query = query.format(session_id)

                order_with_payments = [(int(x.get_entry(0)),
                                        int(x.get_entry(1)),
                                        datetime.strptime(x.get_entry(2), date_format),
                                        float(x.get_entry(3)),
                                        float(x.get_entry(4)),
                                        x.get_entry(5),
                                        x.get_entry(6),
                                        x.get_entry(7),
                                        x.get_entry(8),
                                        x.get_entry(9),
                                        x.get_entry(10)) for x in conn.select(query)]

                all_orders = []  # type: List[Order]
                for order_tuple in order_with_payments:
                    order_id = order_tuple[0]
                    order_state = order_tuple[1]
                    order_date = order_tuple[2]
                    order_total = order_tuple[3]
                    donation_value = order_tuple[4]
                    void_reason_id = order_tuple[5]
                    discount_value = float(order_tuple[6]) if order_tuple[6] else None
                    tip_value = float(order_tuple[7]) if order_tuple[7] else None
                    sale_type = int(order_tuple[8])
                    table_id = order_tuple[9]
                    delivery_originator = order_tuple[10]

                    order = Order(order_id, order_date, order_state, order_total, donation_value, [], void_reason_id, discount_value, tip_value, sale_type, table_id, None, delivery_originator)
                    all_orders.append(order)

                return all_orders
            except:
                return list({"error": "error"}.values())

        report_pos_list = self.pos_list if report_pos is None else (report_pos, )
        return self.execute_parallel_in_all_databases_returning_flat_list(inner_func, report_pos_list)

    def _get_order_from_xml(self, xml_file_name, xml_content):
        if xml_content.find("<CFe>") >= 0:
            return self._get_order_from_sat_xml(xml_file_name, xml_content)
        elif xml_content.find("<NFe") >= 0:
            return self._get_order_from_nfce_xml(xml_content)
        else:
            return None

    def _get_order_from_sat_xml(self, xml_file_name, xml_content):
        parsed_xml = eTree.XML(xml_content)
        index1 = xml_file_name.find("_")
        index2 = xml_file_name.find("_", index1 + 1)
        order_id = int(xml_file_name[index1 + 1: index2])

        nf_number = parsed_xml.find("infCFe/ide/nCFe").text
        data_nota = parsed_xml.find("infCFe/ide/dEmi").text
        hora_nota = parsed_xml.find("infCFe/ide/hEmi").text
        data_emissao = datetime.strptime(data_nota + hora_nota, "%Y%m%d%H%M%S")

        v_desc = float(parsed_xml.find("infCFe/total/ICMSTot/vDesc").text)
        v_total = round(float(parsed_xml.find("infCFe/total/ICMSTot/vProd").text) - v_desc, 2)

        tenders = []  # type: List[OrderTender]
        troco = parsed_xml.find("infCFe/pgto/vTroco")
        pgtos = parsed_xml.findall("infCFe/pgto/MP")
        for pgto in pgtos:
            c_mp = int(pgto.find("cMP").text)
            v_mp = float(pgto.find("vMP").text)

            if c_mp == 1:  # DINHEIRO
                tender_type = 0
            elif c_mp == 3:  # CREDITO:
                tender_type = 1
            elif c_mp == 4:  # DEBITO
                tender_type = 2
            else:
                tender_type = 99

            if troco is not None:
                v_mp = round(v_mp - float(troco.text), 2)
                troco = None

            tender = OrderTender(tender_type, v_mp)
            tenders.append(tender)

        order = Order(order_id, data_emissao, 5, v_total, 0, tenders, None, v_desc, None, None, None, None)
        order.nf_number = nf_number
        return order

    def _get_order_from_nfce_xml(self, xml_content):
        nfe_namespace = "http://www.portalfiscal.inf.br/nfe"

        parsed_xml = eTree.XML(xml_content)
        c_nf = parsed_xml.find("{{{0}}}NFe/{{{0}}}infNFe/{{{0}}}ide/{{{0}}}cNF".format(nfe_namespace)).text
        nf_number = parsed_xml.find("{{{0}}}NFe/{{{0}}}infNFe/{{{0}}}ide/{{{0}}}nNF".format(nfe_namespace)).text
        n_serie = parsed_xml.find("{{{0}}}NFe/{{{0}}}infNFe/{{{0}}}ide/{{{0}}}serie".format(nfe_namespace)).text
        order_id = int(n_serie + c_nf)

        data_hora_nota = parsed_xml.find("{{{0}}}NFe/{{{0}}}infNFe/{{{0}}}ide/{{{0}}}dhEmi".format(nfe_namespace)).text
        data_emissao = iso8601.parse_date(data_hora_nota)

        v_desc = float(parsed_xml.find("{{{0}}}NFe/{{{0}}}infNFe/{{{0}}}total/{{{0}}}ICMSTot/{{{0}}}vDesc".format(nfe_namespace)).text)
        v_total = round(float(parsed_xml.find("{{{0}}}NFe/{{{0}}}infNFe/{{{0}}}total/{{{0}}}ICMSTot/{{{0}}}vProd".format(nfe_namespace)).text) - v_desc, 2)

        tenders = []  # type: List[OrderTender]
        pgtos = parsed_xml.findall("{{{0}}}NFe/{{{0}}}infNFe/{{{0}}}pag".format(nfe_namespace))[0]
        v_troco = pgtos.find(".//{{{0}}}vTroco".format(nfe_namespace))
        v_troco = float(v_troco.text) if v_troco is not None else 0.0

        for pgto in pgtos:
            if "vTroco" in pgto.tag:
                continue

            t_pag = int(pgto.find(".//{{{0}}}tPag".format(nfe_namespace)).text)
            v_pag = float(pgto.find(".//{{{0}}}vPag".format(nfe_namespace)).text)

            if t_pag == 1:  # DINHEIRO
                tender_type = 0
            elif t_pag == 3:  # CREDITO:
                tender_type = 1
            elif t_pag == 4:  # DEBITO
                tender_type = 2
            else:
                tender_type = 99

            if v_troco != 0.0:
                v_pag = round(v_pag - v_troco, 2)
                v_troco = 0.0

            tender = OrderTender(tender_type, v_pag)
            tenders.append(tender)

        order = Order(order_id, data_emissao, 5, v_total, 0, tenders, None, v_desc, None, None, None, None)
        order.nf_number = nf_number
        return order

    def get_payments_by_order(self, paid_orders):
        def inner_func(conn):
            order_list = map(lambda order: order.order_id, paid_orders)
            order_list = ','.join(map(str, order_list))
            query = _GetPaymentsByOrders
            query = query.format(order_list)

            return [(int(x.get_entry(0)), float(x.get_entry(1)), int(x.get_entry(2))) for x in conn.select(query)]

        ret = self.execute_parallel_in_all_databases_returning_flat_list(inner_func, self.pos_list)

        grouped_payments_by_order = {}
        for tender_type in ret:
            tender_id = tender_type[0]
            tender_grouped_value = tender_type[1]
            tender_grouped_quantity = tender_type[2]

            if tender_id not in grouped_payments_by_order:
                grouped_payments_by_order[tender_id] = {'value': tender_grouped_value,
                                                        'quantity': tender_grouped_quantity,
                                                        'id': tender_grouped_quantity}
            else:
                grouped_payments_by_order[tender_id]['value'] += tender_grouped_value
                grouped_payments_by_order[tender_id]['quantity'] += tender_grouped_quantity

        return grouped_payments_by_order

    def get_change_sum(self, initial_date, end_date):
        def inner_func(conn):
            query = _GetChangeAmountSum.format(initial_date, end_date)
            return [(x.get_entry(0), float(x.get_entry(1))) for x in conn.select(query)]

        return self.execute_parallel_in_all_databases_returning_flat_list(inner_func, self.pos_list)

    def get_tender_details(self, paid_orders):
        def inner_func(conn):
            all_orders_id = ",".join([str(x.order_id) for x in paid_orders])
            query = """
                select tenderid, 
                       tenderamount, 
                       tenderdetail, 
                       orderid 
                 from ordertender 
                 where orderid in ({})""".format(all_orders_id)

            return [(x.get_entry("TenderId"),
                     x.get_entry("TenderAmount"),
                     x.get_entry("TenderDetail"),
                     x.get_entry("OrderId")) for x in conn.select(query)]

        return self.execute_parallel_in_all_databases_returning_flat_list(inner_func, self.pos_list)

    def get_opened_orders(self, report_type, session_id=None, initial_date=None, end_date=None, operator_id=None):
        def inner_func(conn):
            query = """
            SELECT DISTINCT(o.OrderId) 
                FROM Orders o 
                JOIN ServiceOrders so ON so.OrderId == o.OrderId
                LEFT JOIN TableService ts on so.ServiceId = ts.ServiceId
                LEFT JOIN ServiceSplit sp on so.OrderId = sp.SplitOrderId
            WHERE sp.SplitOrderId IS NULL AND o.StateId = {}
              AND ts.FinishedTS IS NULL""".format(MwOrderStatus.STORED.value)

            if operator_id:
                query += " and so.UserId = {}".format(operator_id)

            if report_type == self.TypeSessionId:
                if not session_id:
                    return []
                query += " and SessionId like '%{}%';".format(session_id)
            else:
                if not initial_date or not end_date:
                    return []
                date_format = "%Y%m%d"
                formatted_initial_date = initial_date.strftime(date_format)
                formatted_end_date = end_date.strftime(date_format)
                if report_type == self.TypeBusinessPeriod:
                    query += " and substr(SessionID, instr(SessionID, 'period=')+7) >= '{0}' " \
                             " and substr(SessionID, instr(SessionID, 'period=')+7) <= '{1}';"\
                        .format(formatted_initial_date, formatted_end_date)
                elif report_type == self.TypeRealDate:
                    query += " and strftime('%Y%m%d', CreatedAt, 'localtime') >= '{}' " \
                             " and strftime('%Y%m%d', CreatedAt, 'localtime') <= '{}';"\
                        .format(formatted_initial_date, formatted_end_date)

            return [row.get_entry("OrderId") for row in conn.select(query)]

        return self.execute_parallel_in_all_databases_returning_flat_list(inner_func, self.pos_list)

    _PaidOrdersByRealDateQuery = \
"""select o.OrderId, o.StateId, od.OrderDate, o.TotalGross, ot.TenderId, sum(ot.TenderAmount - coalesce(ot.ChangeAmount, 0)), coalesce(ocp.Value, 0), o.DiscountAmount, o.Tip, o.SaleType, tab.value, coalesce(dly.Value, 'Delivery')
    from Orders o
    inner join OrderTender ot on o.OrderId = ot.OrderId
    inner join (
        select OrderId, strftime('%Y-%m-%d', Value, 'localtime') OrderDate
        from OrderCustomProperties
        where key = 'FISCALIZATION_DATE' 
            and strftime('%Y-%m-%d', Value, 'localtime') >= '{0}' 
            and strftime('%Y-%m-%d', Value, 'localtime') <= '{1}'
    ) od on o.OrderId = od.OrderId
    left join OrderCustomProperties ocp on ocp.OrderId = o.OrderId and ocp.Key = 'DONATION_VALUE'
    left join OrderCustomProperties tab on tab.OrderId = o.OrderId and tab.Key = 'TABLE_ID'
    left join OrderCustomProperties dly on dly.OrderId = o.OrderId and dly.Key = 'BRAND'
    where o.StateId = 5 and SessionId like '%{2}%'
    group by ot.orderid"""

    _VoidedOrdersByRealDateQuery = \
"""select o.OrderId, o.StateId, od.OrderDate, coalesce(o.TotalGross, 0), coalesce(ocp.Value, 0), vrs.Value as VoidReason, o.DiscountAmount, o.Tip, o.SaleType, tab.value, coalesce(dly.Value, 'Delivery')
    from Orders o
    inner join (
        select OrderId, OrderDate 
        from (
            select OrderId, strftime('%Y-%m-%d', max(Timestamp) , 'localtime') OrderDate
            from (
                select OrderId, StateId, Timestamp from OrderStateHistory 
                where strftime('%Y-%m-%d', Timestamp, 'localtime') > strftime('%Y-%m-%d', '{0}' , '-1 day') 
                  and strftime('%Y-%m-%d', Timestamp, 'localtime') < strftime('%Y-%m-%d', '{1}' , '+1 day')
            )
            where StateId = 4
            group by OrderId
        ) a where a.OrderDate >= '{0}' and a.OrderDate <= '{1}'
    ) od on o.OrderId = od.OrderId
    left join OrderCustomProperties vrs on vrs.OrderId = o.OrderId and vrs.Key = 'VOID_REASON_ID'
    left join OrderCustomProperties ocp on ocp.OrderId = o.OrderId and ocp.Key = 'DONATION_VALUE'
    left join OrderCustomProperties tab on tab.OrderId = o.OrderId and tab.Key = 'TABLE_ID'
    left join OrderCustomProperties dly on dly.OrderId = o.OrderId and dly.Key = 'BRAND'
    where o.StateId = 4 and SessionId like '%{2}%'
    group by o.orderid"""

    _PaidOrderByBusinessPeriod = \
"""select o.OrderId, o.StateId, substr(o.SessionID, instr(o.SessionID, 'period=')+7), o.TotalGross, ot.TenderId, sum(ot.TenderAmount - coalesce(ot.ChangeAmount, 0)), coalesce(ocp.Value, 0), o.DiscountAmount, o.Tip, o.SaleType, tab.value, coalesce(dly.Value, 'Delivery')
    from Orders o
    inner join OrderTender ot on o.OrderId = ot.OrderId
    left join OrderCustomProperties ocp on ocp.OrderId = o.OrderId and ocp.Key = 'DONATION_VALUE'
    left join OrderCustomProperties tab on tab.OrderId = o.OrderId and tab.Key = 'TABLE_ID'
    left join OrderCustomProperties dly on dly.OrderId = o.OrderId and dly.Key = 'BRAND'
    where o.StateId = 5 and SessionId like '%{2}%'
        and substr(o.SessionID, instr(o.SessionID, 'period=')+7) >= '{0}'
        and substr(o.SessionID, instr(o.SessionID, 'period=')+7) <= '{1}'
    group by o.orderid"""

    _VoidedOrdersByBusinessPeriod = \
"""select o.OrderId, o.StateId, substr(o.SessionID, instr(o.SessionID, 'period=')+7), coalesce(o.TotalGross, 0), coalesce(ocp.Value, 0), vrs.Value, o.DiscountAmount, o.Tip, o.SaleType, tab.value, coalesce(dly.Value, 'Delivery')
    from Orders o
    left join OrderCustomProperties vrs on vrs.OrderId = o.OrderId and vrs.Key = 'VOID_REASON_ID'
    left join OrderCustomProperties ocp on ocp.OrderId = o.OrderId and ocp.Key = 'DONATION_VALUE'
    left join OrderCustomProperties tab on tab.OrderId = o.OrderId and tab.Key = 'TABLE_ID'
    left join OrderCustomProperties dly on dly.OrderId = o.OrderId and dly.Key = 'BRAND'
    where o.StateId = 4 and SessionId like '%{2}%'
        and substr(o.SessionID, instr(o.SessionID, 'period=')+7) >= '{0}'
        and substr(o.SessionID, instr(o.SessionID, 'period=')+7) <= '{1}'
    group by o.orderid"""

    _PaidOrderBySessionId = \
"""select o.OrderId, o.StateId, o.BusinessPeriod, o.TotalGross, ot.TenderId, sum(ot.TenderAmount - coalesce(ot.ChangeAmount, 0)), coalesce(ocp.Value, 0), o.DiscountAmount, o.Tip, o.SaleType, tab.value, coalesce(dly.Value, 'Delivery')
    from Orders o
    inner join OrderTender ot on o.OrderId = ot.OrderId
    left join OrderCustomProperties ocp on ocp.OrderId = o.OrderId and ocp.Key = 'DONATION_VALUE'
    left join OrderCustomProperties tab on tab.OrderId = o.OrderId and tab.Key = 'TABLE_ID'
    left join OrderCustomProperties dly on dly.OrderId = o.OrderId and dly.Key = 'BRAND'
    where o.SessionId = '{0}' and o.StateId = 5
    group by o.orderid"""

    _VoidedOrdersBySessionId = \
"""select o.OrderId, o.StateId, o.BusinessPeriod, coalesce(o.TotalGross, 0), coalesce(ocp.Value, 0), vrs.Value as VoidReason, o.DiscountAmount, o.Tip, o.SaleType, tab.value, coalesce(dly.Value, 'Delivery')
    from Orders o
    left join OrderCustomProperties ocp on ocp.OrderId = o.OrderId and ocp.Key = 'DONATION_VALUE'
    left join OrderCustomProperties vrs on vrs.OrderId = o.OrderId and vrs.Key = 'VOID_REASON_ID'
    left join OrderCustomProperties tab on tab.OrderId = o.OrderId and tab.Key = 'TABLE_ID'
    left join OrderCustomProperties dly on dly.OrderId = o.OrderId and dly.Key = 'BRAND'
    where o.SessionId = '{0}' and o.StateId = 4
    group by o.orderid"""

    _QueryHourlySale = """
SELECT storetype, 
       orderdate, 
       hora, 
       minuto, 
       SUM(totalgross) total,
	   count(1) total_transcations
FROM   (SELECT ( CASE 
                   WHEN distributionpoint == 'DT' THEN 'DT' 
                   ELSE CASE WHEN distributionpoint == 'KK' THEN 'KK'
				   ELSE 'STORE' 
				  END 
                 END )                                      StoreType, 
               Strftime('%Y-%m-%d', orderdate, 'localtime') OrderDate, 
               Strftime('%H', orderdate, 'localtime')       HORA, 
               ( CASE 
                   WHEN Cast(Strftime('%M', orderdate, 'localtime')AS INT) > 29 
                 THEN 
                   '30' 
                   ELSE '00' 
                 END )                                      AS MINUTO, 
               o.totalgross 
        FROM   orders o 
               inner join ordertender ot 
                       ON o.orderid = ot.orderid 
               inner join ( 
                          -- Busca as Orders do dia que estamos interessadosh  
                          SELECT orderid, 
                                 value OrderDate 
                           FROM   ordercustomproperties 
                           WHERE  KEY = 'FISCALIZATION_DATE' 
                                  AND Strftime('%Y-%m-%d', value, 'localtime') 
                                      >= 
                                      '{}' 
                                  AND Strftime('%Y-%m-%d', value, 'localtime') 
                                      <= 
                                      '{}') 
                                                            od 
                       ON o.orderid = od.orderid 
               left join ordercustomproperties ocp 
                      ON ocp.orderid = o.orderid 
                         AND ocp.KEY = 'DONATION_VALUE' 
        WHERE  o.stateid = 5 and o.SessionId like '%user={},%') a 
GROUP  BY storetype, 
          orderdate, 
          hora, 
          minuto  """


_GetPaymentsByOrders = \
"""SELECT TenderId, SUM(COALESCE(TenderAmount, 0) - COALESCE(ChangeAmount, 0)) as Amount, COUNT(OrderId)
FROM OrderTender
WHERE OrderId in ({0})
GROUP BY TenderId"""


_GetChangeAmountSum = \
"""select o.sessionid, sum(ot.changeamount)
from orders o
join ordertender ot on o.orderid = ot.orderid
join ordercustomproperties ocp on ocp.orderid = o.orderid
where ot.changeamount is not null
    and key = 'FISCALIZATION_DATE' and strftime('%Y%m%d', Value, 'localtime') >= '{0}' 
    and strftime('%Y%m%d', Value, 'localtime') <= '{1}'
group by o.sessionid
"""



from ._Part import Part
from ._Report import Report
from ._Formatter import Formatter
from ._Generator import Generator
from ._Presenter import Presenter

from ._Reporter import Reporter
from ._ReportJSONEncoder import ReportJSONEncoder
from ._ReportParser import ReportParser



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

"""Module: barcode.codex

:Provided barcodes: Code 39, Code 128, PZN
"""
__docformat__ = 'restructuredtext en'

from barcode.base import Barcode
from barcode.charsets import code128, code39
from barcode.errors import *


# Sizes
MIN_SIZE = 0.2
MIN_QUIET_ZONE = 2.54


def check_code(code, name, allowed):
    wrong = []
    for char in code:
        if char not in allowed:
            wrong.append(char)
    if wrong:
        raise IllegalCharacterError('The following characters are not '
            'valid for {name}: {wrong}'.format(name=name,
                wrong=', '.join(wrong)))


class Code39(Barcode):
    """Initializes a new Code39 instance.

    :parameters:
        code : String
            Code 39 string without \* and checksum (added automatically if
            `add_checksum` is True).
        writer : barcode.writer Instance
            The writer to render the barcode (default: SVGWriter).
        add_checksum : Boolean
            Add the checksum to code or not (default: True).
    """

    name = 'Code 39'

    def __init__(self, code, writer=None, add_checksum=True):
        self.code = code.upper()
        if add_checksum:
            self.code += self.calculate_checksum()
        self.writer = writer or Barcode.default_writer()
        check_code(self.code, self.name, code39.REF)

    def __unicode__(self):
        return self.code

    __str__ = __unicode__

    def get_fullcode(self):
        return self.code

    def calculate_checksum(self):
        check = sum([code39.MAP[x][0] for x in self.code]) % 43
        for k, v in code39.MAP.items():
            if check == v[0]:
                return k

    def build(self):
        chars = [code39.EDGE]
        for char in self.code:
            chars.append(code39.MAP[char][1])
        chars.append(code39.EDGE)
        return [code39.MIDDLE.join(chars)]

    def render(self, writer_options):
        options = dict(module_width=MIN_SIZE, quiet_zone=MIN_QUIET_ZONE)
        options.update(writer_options or {})
        return Barcode.render(self, options)


class PZN(Code39):
    """Initializes new German number for pharmaceutical products.

    :parameters:
        pzn : String
            Code to render.
        writer : barcode.writer Instance
            The writer to render the barcode (default: SVGWriter).
    """

    name = 'Pharmazentralnummer'

    digits = 6

    def __init__(self, pzn, writer=None):
        pzn = pzn[:self.digits]
        if not pzn.isdigit():
            raise IllegalCharacterError('PZN can only contain numbers.')
        if len(pzn) != self.digits:
            raise NumberOfDigitsError('PZN must have {0} digits, not '
                                      '{1}.'.format(self.digits, len(pzn)))
        self.pzn = pzn
        self.pzn = '{0}{1}'.format(pzn, self.calculate_checksum())
        Code39.__init__(self, 'PZN-{0}'.format(self.pzn), writer,
                        add_checksum=False)

    def get_fullcode(self):
        return 'PZN-{0}'.format(self.pzn)

    def calculate_checksum(self):
        sum_ = sum([int(x) * int(y) for x, y in enumerate(self.pzn, start=2)])
        checksum = sum_ % 11
        if checksum == 10:
            raise BarcodeError('Checksum can not be 10 for PZN.')
        else:
            return checksum


class Code128(Barcode):
    """Initializes a new Code128 instance. The checksum is added automatically
    when building the bars.

    :parameters:
        code : String
            Code 128 string without checksum (added automatically).
        writer : barcode.writer Instance
            The writer to render the barcode (default: SVGWriter).
    """

    name = 'Code 128'

    def __init__(self, code, writer=None):
        self.code = code
        self.writer = writer or Barcode.default_writer()
        self._charset = 'B'
        self._buffer = ''
        check_code(self.code, self.name, code128.ALL)

    def __unicode__(self):
        return self.code

    __str__ = __unicode__

    @property
    def encoded(self):
        return self._build()

    def get_fullcode(self):
        return self.code

    def _new_charset(self, which):
        if which == 'A':
            code = self._convert('TO_A')
        elif which == 'B':
            code = self._convert('TO_B')
        elif which == 'C':
            code = self._convert('TO_C')
        self._charset = which
        return [code]

    def _maybe_switch_charset(self, pos):
        char = self.code[pos]
        next_ = self.code[pos:pos + 10]

        def look_next():
            digits = 0
            for c in next_:
                if c.isdigit():
                    digits += 1
                else:
                    break
            return digits > 3

        codes = []
        if self._charset == 'C' and not char.isdigit():
            if char in code128.B:
                codes = self._new_charset('B')
            elif char in code128.A:
                codes = self._new_charset('A')
            if len(self._buffer) == 1:
                codes.append(self._convert(self._buffer[0]))
                self._buffer = ''
        elif self._charset == 'B':
            if look_next():
                codes = self._new_charset('C')
            elif char not in code128.B:
                if char in code128.A:
                    codes = self._new_charset('A')
        elif self._charset == 'A':
            if look_next():
                codes = self._new_charset('C')
            elif char not in code128.A:
                if char in code128.B:
                    codes = self._new_charset('B')
        return codes

    def _convert(self, char):
        if self._charset == 'A':
            return code128.A[char]
        elif self._charset == 'B':
            return code128.B[char]
        elif self._charset == 'C':
            if char in code128.C:
                return code128.C[char]
            elif char.isdigit():
                self._buffer += char
                if len(self._buffer) == 2:
                    value = int(self._buffer)
                    self._buffer = ''
                    return value

    def _try_to_optimize(self, encoded):
        if encoded[1] in code128.TO:
            encoded[:2] = [code128.TO[encoded[1]]]
        return encoded

    def _calculate_checksum(self, encoded):
        cs = [encoded[0]]
        for i, code_num in enumerate(encoded[1:], start=1):
            cs.append(i * code_num)
        return sum(cs) % 103

    def _build(self):
        encoded = [code128.START_CODES[self._charset]]
        for i, char in enumerate(self.code):
            encoded.extend(self._maybe_switch_charset(i))
            code_num = self._convert(char)
            if code_num is not None:
                encoded.append(code_num)
        # Finally look in the buffer
        if len(self._buffer) == 1:
            encoded.extend(self._new_charset('B'))
            encoded.append(self._convert(self._buffer[0]))
            self._buffer = ''
        encoded = self._try_to_optimize(encoded)
        return encoded

    def build(self):
        encoded = self._build()
        encoded.append(self._calculate_checksum(encoded))
        code = ''
        for code_num in encoded:
            code += code128.CODES[code_num]
        code += code128.STOP
        code += '11'
        return [code]

    def render(self, writer_options):
        options = dict(module_width=MIN_SIZE, quiet_zone=MIN_QUIET_ZONE)
        options.update(writer_options or {})
        return Barcode.render(self, options)



class PickListItemInformation(object):
    def __init__(self, qty, default_qty, item_type, name, label, only_flag, min_qty):
        self.qty = qty
        self.default_qty = default_qty
        self.item_type = item_type
        self.name = name
        self.label = label
        self.only_flag = only_flag
        self.min_qty = min_qty



from datetime import datetime  # noqa


class TipReportHeaderDto(object):
    def __init__(self, start_date, end_date, current_date_time, store_id):
        # type: (datetime, datetime, datetime, store_id) -> None

        if start_date > end_date:
            raise Exception("The start_date cannot be greater than end_date")

        self.start_date = start_date  # type: datetime
        self.end_date = end_date  # type: datetime
        self.current_date_time = current_date_time  # type: datetime
        self.store_id = store_id  # type: str



from _OperatorSaleItem import OperatorSaleItem
from _ProductSaleItem import ProductSaleItem
from _SaleChallengeItem import SaleChallengeItem

__all__ = ["OperatorSaleItem", "ProductSaleItem", "SaleChallengeItem"]



# -*- coding: utf-8 -*-

from xml.etree import cElementTree as eTree

from _ModuleLogger import logger
from nfcedisabler import NfceDisablerResponse


class NfceDisablerResponseParser(object):
    def __init__(self, versao_ws):
        self.namespace_soap = "http://www.w3.org/2003/05/soap-envelope"

        if versao_ws == 1:
            self.namespace_ret_inutilizacao = "http://www.portalfiscal.inf.br/nfe/wsdl/NfeInutilizacao2"
        elif versao_ws == 3:
            self.namespace_ret_inutilizacao = "http://www.portalfiscal.inf.br/nfe/wsdl/NfeInutilizacao3"
        else:
            self.namespace_ret_inutilizacao = "http://www.portalfiscal.inf.br/nfe/wsdl/NFeInutilizacao4"

    NAMESPACE_NFE = "http://www.portalfiscal.inf.br/nfe"
    NAMESPACE_NFE_KEY = "{http://www.portalfiscal.inf.br/nfe}"


    def parse_response(self, response_str):
        # type: (str) -> NfceDisablerResponse
        try:
            response_xml = eTree.XML(response_str.encode("utf8"))
            logger.info(response_str)
            logger.info(self.namespace_ret_inutilizacao)
            c_stat = response_xml.findall('.//{0}{1}'.format(self.NAMESPACE_NFE_KEY, 'cStat'))[0].text
            x_motivo = response_xml.findall('.//{0}{1}'.format(self.NAMESPACE_NFE_KEY, 'xMotivo'))[0].text
            n_prot_node = response_xml.findall('.//{0}{1}'.format(self.NAMESPACE_NFE_KEY, 'nProt'))
            n_prot = None
            if n_prot_node is not None and len(n_prot_node) > 0:
                n_prot = n_prot_node[0].text

            return NfceDisablerResponse(c_stat, x_motivo, n_prot)
        except Exception as ex:
            logger.error("Erro parseando XML de response: {0}".format(repr(ex)))



# -*- coding: utf-8 -*-

from datetime import datetime, timedelta

from chatservice.model import Message, MessageFrom
from msgbus import MBEasyContext
from old_helper import BaseRepository
from persistence import Connection
from typing import List


class MessageRepository(BaseRepository):
    DatabaseDateFormat = "%Y-%m-%d %H:%M:%S"
    
    def __init__(self, mb_context):
        # type: (MBEasyContext) -> None
        super(MessageRepository, self).__init__(mb_context)
    
    def get_updates(self, message_id):
        # type: (int) -> List[Message]
        
        def inner_func(conn):
            # type: (Connection) -> List[Message]
            
            query = "SELECT Id, Origin, CreatedTime, ReceivedTime, Text " \
                    "FROM ChatMessages " \
                    "WHERE Id > {0} " \
                    "ORDER BY Id".format(message_id)
            unread_messages_tuples = [(int(x.get_entry(0)),
                                       x.get_entry(1).decode("utf-8"),
                                       x.get_entry(2).decode("utf-8"),
                                       x.get_entry(3).decode("utf-8") if x.get_entry(3) is not None else None,
                                       x.get_entry(4).decode("utf-8"))
                                      for x in conn.select(query)]
            
            unread_messages = self._convert_tuple_to_messages(unread_messages_tuples)
            
            return unread_messages
        
        return self.execute_with_connection(inner_func, service_name="DeliveryPersistence")
    
    def get_last_messages(self, quantity):
        # type: (int) -> List[Message]
        
        def inner_func(conn):
            # type: (Connection) -> List[Message]
            
            query = "SELECT Id, Origin, CreatedTime, ReceivedTime, Text " \
                    "FROM ChatMessages " \
                    "ORDER BY Id DESC " \
                    "LIMIT {0}".format(quantity)
            last_messages_tuples = [(int(x.get_entry(0)),
                                     x.get_entry(1).decode("utf-8"),
                                     x.get_entry(2).decode("utf-8"),
                                     x.get_entry(3).decode("utf-8") if x.get_entry(3) is not None else None,
                                     x.get_entry(4)) for x in conn.select(query)]
            
            return self._convert_tuple_to_messages(last_messages_tuples)
        
        return self.execute_with_connection(inner_func, service_name="DeliveryPersistence")
    
    def save_new_messages_from_store(self, messages):
        # type: (List[Message]) -> List[int]
        
        def inner_func(conn):
            # type: (Connection) -> List[int]
            
            ret = []  # type: List[int]
            for message in messages:
                # noinspection SqlInsertValues
                conn.query("INSERT INTO ChatMessages (Origin, CreatedTime, Text, LastTimeSentToServer) "
                           "VALUES ('{0}', '{1}', '{2}', '{3}')"
                           .format(MessageFrom.STORE,
                                   message.created_time.strftime(self.DatabaseDateFormat),
                                   message.text.replace("'", "''").encode("utf-8"),
                                   datetime.utcnow().strftime(self.DatabaseDateFormat)))
                
                new_id = [int(x.get_entry(0)) for x in conn.select("SELECT LAST_INSERT_ROWID()")][0]
                ret.append(new_id)
            
            return ret
        
        return self.execute_with_transaction(inner_func, service_name="DeliveryPersistence")
    
    def save_new_messages_from_sac(self, messages):
        # type: (List[Message]) -> List[Message]
        
        def inner_func(conn):
            # type: (Connection) -> List[Message]
            
            ret = []  # type: List[Message]
            for message in messages:
                ret_message = Message()
                ret_message.id = message.server_id
                
                select_query = "SELECT ReceivedTime FROM ChatMessages where ServerId = '{0}'".format(message.server_id)
                has_message_on_db = [x.get_entry(0).decode("utf-8") for x in conn.select(select_query)]
                if len(has_message_on_db) == 0:
                    # noinspection SqlInsertValues
                    insert_query = "INSERT INTO ChatMessages (Origin, CreatedTime, ReceivedTime, Text, ServerId) " \
                                   "VALUES ('{0}', '{1}', '{2}', '{3}', '{4}')" \
                        .format(message.message_from,
                                message.created_time.strftime(self.DatabaseDateFormat),
                                message.received_time.strftime(self.DatabaseDateFormat),
                                message.text.encode("utf-8"),
                                message.server_id.encode("utf-8"))
                    
                    conn.query(insert_query)
                    
                    ret_message.received_time = message.received_time
                else:
                    ret_message.received_time = datetime.strptime(has_message_on_db[0], self.DatabaseDateFormat)
                
                ret.append(ret_message)
            
            return ret
        
        return self.execute_with_transaction(inner_func, service_name="DeliveryPersistence")
    
    def mark_messages_received(self, messages):
        # type: (List[Message]) -> None
        
        def inner_func(conn):
            # type: (Connection) -> None
            
            for message in messages:
                conn.query("UPDATE ChatMessages SET ReceivedTime = '{0}' WHERE Id = {1}"
                           .format(message.received_time.strftime(self.DatabaseDateFormat), message.id))
        
        self.execute_with_transaction(inner_func, service_name="DeliveryPersistence")
    
    def get_messages_without_confirmation(self):
        # type: () -> List[Message]
        
        def inner_func(conn):
            # type: (Connection) -> List[Message]
            
            sync_date = datetime.utcnow() - timedelta(seconds=30)
            query = "SELECT Id, CreatedTime, Text " \
                    "FROM ChatMessages " \
                    "WHERE LastTimeSentToServer < '{0}' AND Origin = 'Store' AND ReceivedTime IS NULL " \
                    "ORDER BY Id".format(sync_date.strftime(self.DatabaseDateFormat))
            
            messages_without_confirmation = [(int(x.get_entry(0)),
                                              x.get_entry(1).decode("utf-8"),
                                              x.get_entry(2).decode("utf-8"))
                                             for x in conn.select(query)]
            
            messages = []
            for message_tuple in messages_without_confirmation:
                message = Message()
                message.id = message_tuple[0]
                message.created_time = datetime.strptime(message_tuple[1], self.DatabaseDateFormat)
                message.text = message_tuple[2]
                
                messages.append(message)
            
            return messages
        
        return self.execute_with_connection(inner_func, service_name="DeliveryPersistence")
    
    def _convert_tuple_to_messages(self, messages_tuples):
        # type: (List[tuple]) -> List[Message]
        
        messages = []
        for message_tuple in messages_tuples:
            unread_message = Message()
            unread_message.id = message_tuple[0]
            unread_message.message_from = message_tuple[1]
            unread_message.created_time = datetime.strptime(message_tuple[2], self.DatabaseDateFormat)
            unread_message.received_time = datetime.strptime(message_tuple[3], self.DatabaseDateFormat) \
                if message_tuple[3] is not None else None
            unread_message.text = message_tuple[4]
            
            messages.append(unread_message)
        
        return sorted(messages, key=lambda x: x.id)



from logging import Logger

from interactor.route.menu.get_product_info.dto import GetProductInfoRequest
from model.config import Config
from model.error_type import PosValidationErrorsType
from model.exception import ValidationError
from model.repository import PosRepositoryModel
from model.services import ProductInfoServiceModel
from validation_util import validate_api_key


class GetProductInfoInteractor(object):

    def __init__(
        self,
        logger,                 # type: Logger
        config,                 # type: Config
        pos_repository,         # type: PosRepositoryModel
        product_info_service    # type: ProductInfoServiceModel
    ):
        # type: (...) -> None

        self.logger = logger
        self.config = config
        self.pos_repository = pos_repository
        self.product_info_service = product_info_service

    def execute(
        self,
        request     # type: GetProductInfoRequest
    ):
        # type: (...) -> str

        validate_api_key(
            request=request,
            secure_api_key_list=self.config.api_key_mapping
        )

        product_info = self.product_info_service.get_product_info(language=request.language)

        if not product_info:
            raise ValidationError(PosValidationErrorsType.ProductInfoNotFound)

        return product_info



from httpmessageprocessorutil.automapped import auto_map


@auto_map({"system_api_key": {"name": "X-Api-Key", "location": "header", "type": str},
           "pos_id": {"name": "posId", "type": str},
           "tender_id": {"name": "tenderId", "type": int}})
class CancelPaymentRequest(object):
    def __init__(
            self,
            system_api_key,   # type: str
            pos_id,           # type: str
            tender_id         # type: int
     ):
        # type: (...) -> None

        self.system_api_key = system_api_key
        self.pos_id = pos_id
        self.tender_id = tender_id



# -*- coding: utf-8 -*-
import logging
import bustoken
import re
import subprocess
from threading import Thread, Condition
from msgbus import MBEasyContext, MBException, FM_PARAM, TK_SYS_ACK

logger = logging.getLogger("KdsMonitor")


class Kds:
    def __init__(self, name, ip, view_name):
        # type: (str, str) -> None
        self.name = name
        self.ip = ip
        self.view_name = view_name
        self.error_count = 0
        self.enabled = False


class KdsMonitor(Thread):
    def __init__(self, mbcontext, kdss_to_monitor, monitor_interval, max_error_count, start_wait_time):
        # type: (MBEasyContext, list, int) -> None
        super(KdsMonitor, self).__init__()

        self.mbcontext = mbcontext
        self.kdss_to_monitor = kdss_to_monitor
        self.monitor_interval = monitor_interval
        self.max_error_count = max_error_count
        self.start_wait_time = start_wait_time

        self.running = True
        self.thread_condition = Condition()
        self.ping_regex = re.compile(r'((tempo)|(time))([=<])(.*)ms')

    def run(self):
        self._thread_sleep(self.start_wait_time)
        while self.running:
            try:
                for kds in self.kdss_to_monitor:  # type: Kds
                    logger.info("Monitoring KDS %s" % kds.view_name)
                    if not self._ping(kds.ip):
                        if kds.error_count < self.max_error_count:
                            kds.error_count += 1
                            continue
                        logger.info("Disable KDS %s" % kds.view_name)
                        self._disable_kds(kds)
                    else:
                        if kds.error_count > 0:
                            kds.error_count -= 1
                            continue
                        logger.info("Enable KDS %s" % kds.view_name)
                        self._enable_kds(kds)
            except Exception:
                logger.exception("Erro durante monitoramento dos KDSs")
            finally:
                self._thread_sleep(self.monitor_interval)

    def finish(self):
        with self.thread_condition:
            self.running = False
            self.thread_condition.notifyAll()

    def _disable_kds(self, kds):
        # noinspection PyBroadException
        try:
            # Timeout = 5 Seconds
            ret = self.mbcontext.MB_EasySendMessage("ProductionSystem", bustoken.TK_PROD_DISABLEVIEW, format=FM_PARAM, data="%s" % kds.view_name, timeout=5000 * 1000)
            if ret.token == TK_SYS_ACK:
                kds.enabled = False
        except MBException:
            pass
        except Exception:
            logger.exception("Erro enviando mensagem para desabilitar KDS")

    def _enable_kds(self, kds):
        # noinspection PyBroadException
        try:
            # Timeout = 5 Seconds
            ret = self.mbcontext.MB_EasySendMessage("ProductionSystem", bustoken.TK_PROD_ENABLEVIEW, format=FM_PARAM, data="%s" % kds.view_name, timeout=5000 * 1000)

            if ret.token == TK_SYS_ACK:
                kds.enabled = True
        except MBException:
            pass
        except Exception:
            logger.exception("Erro enviando mensagem para habilitar KDS")

    def _thread_sleep(self, time):
        self.thread_condition.acquire()
        self.thread_condition.wait(time)
        self.thread_condition.release()

    def _ping(self, host):
        import platform

        command = ["ping"]
        params = ["-n", "1"] if platform.system().lower() == "windows" else ["-c", "1"]
        command.extend(params)
        command.append(str(host))
        ping_response = ""

        si = None
        if platform.system().lower() == "windows":
            si = subprocess.STARTUPINFO()
            si.dwFlags |= subprocess.STARTF_USESHOWWINDOW
        ping = subprocess.Popen(command, shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE, startupinfo=si)
        for line in ping.stdout.readlines():
            ping_response += line

        ping.kill()
        if self.ping_regex.search(ping_response) is None:
            return False
        return True



from _StatusResponse import StatusResponse


class GetStoreResponse(object):
    def __init__(
            self,
            status,      # type: bool
            menu_id,     # type: str
            printer,     # type: StatusResponse
            eft,         # type: StatusResponse
            fiscal,      # type: StatusResponse
            manager,     # type: StatusResponse
    ):
        # type: (...) -> None

        self.isOnline = status
        self.menuId = menu_id

        self.printer = printer
        self.eft = eft
        self.fiscal = fiscal
        self.manager = manager





from httpmessageprocessorutil.automapped import auto_map
from typing import List


class ToChangeItems(object):
    def __init__(
            self,
            context,       # type: str
            product_code,  # type: int
            new_quantity,  # type: int
            line_number    # type: int
    ):
        # type: (...) -> None

        self.context = str(context)
        self.product_code = int(product_code)
        self.new_quantity = int(new_quantity)
        self.line_number = int(line_number)


@auto_map({"system_api_key": {"name": "X-Api-Key", "location": "header", "type": str},
           "pos_id": {"name": "posId", "type": str},
           "to_change_items": {"name": "toChangeItems", "type": List[ToChangeItems]}})
class ChangeItemsQuantityRequest(object):
    def __init__(self, system_api_key, pos_id, to_change_items):
        # type: (str, str, List[ToChangeItems]) -> None

        self.system_api_key = system_api_key
        self.pos_id = pos_id
        self.items = to_change_items



# defusedxml
#
# Copyright (c) 2013 by Christian Heimes <christian@python.org>
# Licensed to PSF under a Contributor Agreement.
# See https://www.python.org/psf/license for licensing details.
"""Defuse XML bomb denial of service vulnerabilities
"""
from __future__ import print_function, absolute_import

from .common import (
    DefusedXmlException,
    DTDForbidden,
    EntitiesForbidden,
    ExternalReferenceForbidden,
    NotSupportedError,
    _apply_defusing,
)


def defuse_stdlib():
    """Monkey patch and defuse all stdlib packages

    :warning: The monkey patch is an EXPERIMETNAL feature.
    """
    defused = {}

    from . import cElementTree
    from . import ElementTree
    from . import minidom
    from . import pulldom
    from . import sax
    from . import expatbuilder
    from . import expatreader
    from . import xmlrpc

    xmlrpc.monkey_patch()
    defused[xmlrpc] = None

    for defused_mod in [
        cElementTree,
        ElementTree,
        minidom,
        pulldom,
        sax,
        expatbuilder,
        expatreader,
    ]:
        stdlib_mod = _apply_defusing(defused_mod)
        defused[defused_mod] = stdlib_mod

    return defused


__version__ = "0.6.0"

__all__ = [
    "DefusedXmlException",
    "DTDForbidden",
    "EntitiesForbidden",
    "ExternalReferenceForbidden",
    "NotSupportedError",
]



import sys
import warnings

from six import PY3, binary_type, text_type

from cryptography.hazmat.bindings.openssl.binding import Binding


binding = Binding()
binding.init_static_locks()
ffi = binding.ffi
lib = binding.lib


# This is a special CFFI allocator that does not bother to zero its memory
# after allocation. This has vastly better performance on large allocations and
# so should be used whenever we don't need the memory zeroed out.
no_zero_allocator = ffi.new_allocator(should_clear_after_alloc=False)


def text(charp):
    """
    Get a native string type representing of the given CFFI ``char*`` object.

    :param charp: A C-style string represented using CFFI.

    :return: :class:`str`
    """
    if not charp:
        return ""
    return native(ffi.string(charp))


def exception_from_error_queue(exception_type):
    """
    Convert an OpenSSL library failure into a Python exception.

    When a call to the native OpenSSL library fails, this is usually signalled
    by the return value, and an error code is stored in an error queue
    associated with the current thread. The err library provides functions to
    obtain these error codes and textual error messages.
    """
    errors = []

    while True:
        error = lib.ERR_get_error()
        if error == 0:
            break
        errors.append((
            text(lib.ERR_lib_error_string(error)),
            text(lib.ERR_func_error_string(error)),
            text(lib.ERR_reason_error_string(error))))

    raise exception_type(errors)


def make_assert(error):
    """
    Create an assert function that uses :func:`exception_from_error_queue` to
    raise an exception wrapped by *error*.
    """
    def openssl_assert(ok):
        """
        If *ok* is not True, retrieve the error from OpenSSL and raise it.
        """
        if ok is not True:
            exception_from_error_queue(error)

    return openssl_assert


def native(s):
    """
    Convert :py:class:`bytes` or :py:class:`unicode` to the native
    :py:class:`str` type, using UTF-8 encoding if conversion is necessary.

    :raise UnicodeError: The input string is not UTF-8 decodeable.

    :raise TypeError: The input is neither :py:class:`bytes` nor
        :py:class:`unicode`.
    """
    if not isinstance(s, (binary_type, text_type)):
        raise TypeError("%r is neither bytes nor unicode" % s)
    if PY3:
        if isinstance(s, binary_type):
            return s.decode("utf-8")
    else:
        if isinstance(s, text_type):
            return s.encode("utf-8")
    return s


def path_string(s):
    """
    Convert a Python string to a :py:class:`bytes` string identifying the same
    path and which can be passed into an OpenSSL API accepting a filename.

    :param s: An instance of :py:class:`bytes` or :py:class:`unicode`.

    :return: An instance of :py:class:`bytes`.
    """
    if isinstance(s, binary_type):
        return s
    elif isinstance(s, text_type):
        return s.encode(sys.getfilesystemencoding())
    else:
        raise TypeError("Path must be represented as bytes or unicode string")


if PY3:
    def byte_string(s):
        return s.encode("charmap")
else:
    def byte_string(s):
        return s


# A marker object to observe whether some optional arguments are passed any
# value or not.
UNSPECIFIED = object()

_TEXT_WARNING = (
    text_type.__name__ + " for {0} is no longer accepted, use bytes"
)


def text_to_bytes_and_warn(label, obj):
    """
    If ``obj`` is text, emit a warning that it should be bytes instead and try
    to convert it to bytes automatically.

    :param str label: The name of the parameter from which ``obj`` was taken
        (so a developer can easily find the source of the problem and correct
        it).

    :return: If ``obj`` is the text string type, a ``bytes`` object giving the
        UTF-8 encoding of that text is returned.  Otherwise, ``obj`` itself is
        returned.
    """
    if isinstance(obj, text_type):
        warnings.warn(
            _TEXT_WARNING.format(label),
            category=DeprecationWarning,
            stacklevel=3
        )
        return obj.encode('utf-8')
    return obj



from repository import SequencerRepository
from typing import Optional


class SequencerService(object):
    def __init__(self, sequencer_repository):
        # type: (SequencerRepository) -> None

        self.sequencer_repository = sequencer_repository

    def update_fiscal_id_sequence(self, initial_order_id, last_order_id):
        # type: (int, int) -> None

        fiscal_id = self.sequencer_repository.get_fiscal_id()
        if fiscal_id == 0:
            fiscal_id = initial_order_id + last_order_id
            self.sequencer_repository.set_fiscal_id(fiscal_id)
        elif initial_order_id > fiscal_id:
            fiscal_id = initial_order_id
            self.sequencer_repository.set_fiscal_id(fiscal_id)

    def build_next_fiscal_id(self):
        # type: () -> int

        fiscal_id = self.sequencer_repository.increment_fiscal_id()
        return fiscal_id



from helper import ExtendedEnum


class ListenedEvents(ExtendedEnum):

    TABLE_TOTALED = "TABLE_TOTALED"
    ORDER_MODIFIED = "ORDER_MODIFIED"
    ORDER_MODIFIED_VOID = "ORDER_MODIFIED_VOID_ORDER"
    ORDER_MODIFIED_PAID = "ORDER_MODIFIED_PAID"
    ORDER_MODIFIED_TOTAL = "ORDER_MODIFIED_TOTALED"
    ORDER_MODIFIED_VOID_LINE = "ORDER_MODIFIED_VOID_LINE"
    ORDER_MODIFIED_REOPENED = "ORDER_MODIFIED_REOPENED"
    ORDER_MODIFIED_RECALLED = "ORDER_MODIFIED_RECALLED"



from _PickListItem import PickListItem
from _PickListItemInformation import PickListItemInformation
from _PickListDto import PickListDto
from _PickListHeaderDto import PickListHeaderDto
from _PickListBodyDto import PickListBodyDto

__all__ = ["PickListItem",
           "PickListItemInformation",
           "PickListDto",
           "PickListHeaderDto",
           "PickListBodyDto"]



from interactor import BaseInteractor
from messagebus import Message, DefaultToken, DataType
from messageprocessorutil import UuidMessageProcessor
from model.log import LogEntry


class EventProcessor(UuidMessageProcessor):
    def __init__(self, interactor):
        # type: (BaseInteractor) -> None
        super(EventProcessor, self).__init__()
        self.interactor = interactor

    def call_business(self, log_entry):
        # type: (LogEntry) -> None
        self.interactor.execute(log_entry)

    def format_response(self, message_bus, message, event, input_obj, result):
        message_bus.reply_event(event, Message(DefaultToken.TK_SYS_NAK))

    def format_exception(self, message_bus, message, event, input_obj, exception):
        message_bus.reply_event(event, Message(DefaultToken.TK_SYS_NAK))

    def format_parse_exception(self, message_bus, message, event, data, exception):
        message_bus.reply_event(event, Message(DefaultToken.TK_SYS_NAK))



from _LoginSummaryHeaderDto import LoginSummaryHeaderDto  # noqa
from _LoginSummaryBodyDto import LoginSummaryBodyDto  # noqa


class LoginSummaryDto(object):
    def __init__(self, login_summary_header_dto, login_summary_body_dto):
        # type:(LoginSummaryHeaderDto, LoginSummaryBodyDto) -> ()
        self.login_summary_header_dto = login_summary_header_dto
        self.login_summary_body_dto = login_summary_body_dto



# encoding: utf-8
from _TefCancellationBodyDto import TefCancellationBodyDto
from _TefCancellationHeaderDto import TefCancellationHeaderDto


class TefCancellationGeneratorDto(object):
    def __init__(
        self,
        header,     # type: TefCancellationHeaderDto
        body        # type: TefCancellationBodyDto
    ):
        # type: (...) -> None

        self.header = header
        self.body = body



from messagebus import TokenCreator, TokenPriority
from messagehandler import MessageHandlerBuilder
from messageprocessor import MessageProcessorMessageHandler, DefaultMessageProcessorExecutorFactory, \
    LoggingProcessorCallback
from mobileposprinter.processor import PrnPrintProcessor

PERIPHERALS_GROUP = "6"
TK_PRN_PRINT = TokenCreator.create_token(TokenPriority.high, PERIPHERALS_GROUP, "101")


class MobilePosPrinterMessageHandlerBuilder(MessageHandlerBuilder):
    def __init__(self, pos_id, message_bus, logger):
        self.pos_id = pos_id
        self.message_bus = message_bus
        self.logger = logger

        self.message_handler = None

    def build_singletons(self):
        message_processor = {
            TK_PRN_PRINT: PrnPrintProcessor(self.pos_id, self.message_bus)
        }

        callbacks = [LoggingProcessorCallback(self.logger)]

        self.message_handler = MessageProcessorMessageHandler(
            None,
            message_processor,
            None,
            DefaultMessageProcessorExecutorFactory(callbacks),
            self.logger
        )

    def build_message_handler(self):
        return self.message_handler

    def destroy_message_handler(self, message_handler):
        pass

    def destroy_singletons(self):
        pass



import io
import json

from xml.etree import cElementTree as eTree
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)
from datetime import (
    datetime,
    timedelta,
)
from threading import Lock
from time import sleep

import sysactions
from sysactions import (
    set_custom,
    get_user_information,
)
from msgbus import (
    TK_SYS_NAK,
    TK_POS_GETPOSLIST,
    MBEasyContext,
)

from dailygoals_services.model import (
    Goals,
    GoalsCache,
    ItemGoals,
    NotCacheGoals,
    OperatorGoals,
    PartCode,
    StoreGoals,
)
from dailygoals_services.model.repository import (
    DailyGoalsRepository,
    OrderRepository,
    ProductRepository,
)
from externals.util import (
    LoggerUtil,
    contexts_match,
)

logger = LoggerUtil.get_logger_name(__name__)
DATE_FORMAT = "%Y-%m-%d"


class DailyGoalsService(object):
    def __init__(
        self,
        daily_goals_repository,     # type: DailyGoalsRepository
        order_repository,           # type: OrderRepository
        product_repository,         # type: ProductRepository
        mb_context,                 # type: MBEasyContext
    ):
        # type: (...) -> None

        self.repository = daily_goals_repository
        self.order_repository = order_repository
        self.product_repository = product_repository
        self.lock = Lock()
        self.mb_context = mb_context
        sysactions.mbcontext = mb_context

        self.pos_list = None
        self.timestamp = None
        self.daily_goals = None
        self.today_goals = None
        self.store_id = None
        self.goals_cache_filename = None
        self.to_unicode = self._set_unicode()

    @staticmethod
    def _set_unicode():
        try:
            to_unicode = unicode
        except NameError:
            to_unicode = str
        return to_unicode

    def configure(
        self,
        store_id,               # type: str
        goals_cache_filename,   # type: str
    ):
        # type: (...) -> None

        logger.info('[configure] Begin')
        self.store_id = store_id
        self.goals_cache_filename = goals_cache_filename

        try:
            self._get_pos_list()
            if self.update_daily_goals():
                self.update_daily_sold(get_values_from_database=True)
            else:
                for pos_id in self.pos_list:
                    set_custom(
                        posid=pos_id,
                        key='DAILY_GOALS',
                        value='{}',
                    )

        except Exception as _:
            logger.exception('[configure] Exception:')

        logger.info('[configure] End')

    def _get_pos_list(self):
        # type: () -> None

        retry_cnt = 10
        while retry_cnt:
            msg = self.mb_context.MB_EasySendMessage(
                dest_name="PosController",
                token=TK_POS_GETPOSLIST,
            )
            if msg.token == TK_SYS_NAK:
                retry_cnt -= 1
                sleep(1)
            else:
                self.pos_list = sorted(map(int, msg.data.split("\0")))
                retry_cnt = 0

    def update_daily_goals(self):
        # type: () -> Optional[bool]

        logger.info("[update_daily_goals] Begin")
        now = datetime.now()
        cur_date = now.strftime("%d/%m/%Y")

        self.timestamp = cur_date

        try:
            items_goals, store_goals = self._load_goals_from_cache()
            if not store_goals and not items_goals:
                raise NotCacheGoals()

            item_goals = self._populate_item_goals(items_goals)
            store_goals = self._populate_store_goals(store_goals, item_goals)

            with self.lock:
                self.today_goals = GoalsCache(
                    store_goals=store_goals,
                    item_goals=item_goals,
                )
            logger.info('[update_daily_goals] End')
            return True

        except Exception as _:
            logger.exception('[update_daily_goals] Exception. Sort issue')
            raise NotImplementedError()

    def _save_to_file(
        self,
        goals,  # type: GoalsCache
    ):
        # type: (...) -> None

        with io.open(self.goals_cache_filename, 'w', encoding='utf8') as outfile:
            json_goals = json.dumps(
                goals,
                default=lambda o: o.__dict__,
                sort_keys=True
            )
            outfile.write(self.to_unicode(json_goals))

    def _load_goals_from_cache(self):
        # type: () -> (List[Dict], List[Dict])

        with open(self.goals_cache_filename) as data_file:
            logger.info("[update_daily_goals] Taking goals from cache file")

            response_json = json.load(data_file)
            store_goals_json = response_json.get("storeGoals", [])
            items_goals_json = response_json.get("itemGoals", {})
        return items_goals_json, store_goals_json

    def _populate_item_goals(
        self,
        items_goals_json,   # type: List[Dict]
    ):
        # type: (...) -> List[ItemGoals]

        item_goals = []
        if items_goals_json:
            for item in items_goals_json:
                start_date = datetime.strptime(item["startDate"], DATE_FORMAT)
                end_date = datetime.strptime(item["endDate"], DATE_FORMAT) + timedelta(days=1)
                if start_date < datetime.now() < end_date:
                    name = item.get("name")
                    part_codes = self._build_part_codes_objects(part_codes=item.get("partCodes"))
                    part_codes_set = self._build_part_codes_objects(part_codes=item.get("partCodesSet"))
                    quantity = item.get("quantity")
                    quantity_operator_goal = item.get("operatorQuantity")
                    item_goals.append(ItemGoals(
                        name=name,
                        part_codes=part_codes,
                        part_codes_set=part_codes_set,
                        quantity_goal=quantity,
                        quantity_operator_goal=quantity_operator_goal,
                        quantity_sold=0,
                    ))
        return item_goals

    @staticmethod
    def _build_part_codes_objects(
        part_codes,     # type: List
    ):
        # type: (...) -> List

        if not part_codes:
            return []

        part_code_objects = []
        for part_code in part_codes:
            obj = PartCode(
                id=part_code.get("id"),
                context=part_code.get("context"),
            )
            part_code_objects.append(obj)
        return part_code_objects

    @staticmethod
    def _populate_store_goals(
        store_goals_json,   # type: List[Dict]
        item_goals,         # type: List[ItemGoals]
    ):
        # type: (...) -> StoreGoals

        store_goals = None
        if store_goals_json:
            for item in store_goals_json:
                start_date = datetime.strptime(item["startDate"], DATE_FORMAT)
                end_date = datetime.strptime(item["endDate"], DATE_FORMAT) + timedelta(days=1)
                if start_date < datetime.now() < end_date:
                    total_sales = item.get("totalSales")
                    average_sale_value = item.get("averageSaleValue")
                    operator_sale_goal = item.get("operatorSaleValue")
                    store_goals = StoreGoals(
                        amount_goal=total_sales,
                        average_sale_goal=average_sale_value,
                        operator_sale_goal=operator_sale_goal,
                        item_goals=item_goals,
                    )
                    break
        return store_goals

    def update_daily_sold(
        self,
        order_picture='',                   # type: Optional[eTree.Element]
        operator_id='',                     # type: Optional[str]
        number_of_seats=1,                  # type: Optional[int]
        get_values_from_database=False,     # type: Optional[bool]
    ):
        # type: (...) -> None

        if not self.today_goals:
            return

        now = datetime.now()
        cur_date = now.strftime("%d/%m/%Y")
        if self.timestamp != cur_date:
            self.update_daily_goals()
            get_values_from_database = True

        if self.daily_goals is None:
            get_values_from_database = True

        with self.lock:
            if order_picture:
                self.daily_goals = self._update_total_sold_from_mem(
                    order_picture=order_picture,
                    number_of_seats=number_of_seats,
                )
            elif get_values_from_database:
                store_goals = self.today_goals.store_goals
                items_goals = self.today_goals.item_goals
                operators_goals = []

                initial_date = datetime.now()
                end_date = datetime.now()
                self._populate_operators_sold_values(
                    end_date=end_date,
                    initial_date=initial_date,
                    store_goals=store_goals,
                    item_goals=items_goals,
                    operators_goals=operators_goals,
                )
                if store_goals is not None:
                    self._populate_store_sold_values(
                        end_date=end_date,
                        initial_date=initial_date,
                        operator_id=operator_id,
                        store_goals=store_goals,
                    )

                self.daily_goals = Goals(
                    store_goals=store_goals,
                    operator_goals=operators_goals,
                )

            goals_str = json.dumps(
                self.daily_goals,
                default=lambda o: o.__dict__,
                sort_keys=True
            )

            logger.info(goals_str)
        for pos_id in self.pos_list:
            set_custom(
                posid=pos_id,
                key='DAILY_GOALS',
                value=goals_str,
            )

    def _update_total_sold_from_mem(
        self,
        order_picture,         # type: eTree.Element
        number_of_seats,       # type: int
    ):
        # type: (...) -> Optional[Goals]

        operator_id = order_picture.get("sessionId").split("user=")[1].split(",")[0]
        daily_goals = self.daily_goals
        if daily_goals is None:
            return

        if not any(x.id == operator_id for x in daily_goals.operator_goals):
            self._add_new_operator_to_mem(
                daily_goals=daily_goals,
                operator_id=operator_id,
            )

        added_tickets = False
        sale_lines = order_picture.findall("SaleLine")
        for line in sale_lines:
            result = self._process_operator_goal(
                added_tickets=added_tickets,
                daily_goals=daily_goals,
                line=line,
                number_of_seats=number_of_seats,
                operator_id=operator_id,
            )
            added_tickets = result if result else added_tickets

        return daily_goals

    def _process_operator_goal(
        self,
        added_tickets,      # type: bool
        daily_goals,        # type: Goals
        line,               # type: eTree.Element
        number_of_seats,    # type: int
        operator_id,        # type: str
    ):
        # type (...) -> bool

        part_code = line.get("partCode")
        context = line.get("itemId")
        for operator_daily_goal in daily_goals.operator_goals:
            if operator_daily_goal.id != operator_id:
                continue

            for index, daily_goal_item in enumerate(operator_daily_goal.item_goals):
                if not self._check_part_code_context(
                    part_code=part_code,
                    context=context,
                    part_codes=daily_goal_item.part_codes,
                ):
                    continue

                operator_daily_goal = self._update_sold_info(
                    line=line,
                    daily_goal=operator_daily_goal,
                    part_code=part_code,
                    context=context,
                    key="amount_sold",
                    daily_goal_index=index,
                )

            if daily_goals.store_goals:
                for index, store_goal_item in enumerate(daily_goals.store_goals.item_goals):
                    if not self._check_part_code_context(
                        part_code=part_code,
                        context=context,
                        part_codes=store_goal_item.part_codes,
                    ):
                        continue

                    daily_goals.store_goals = self._update_sold_info(
                        line=line,
                        daily_goal=daily_goals.store_goals,
                        part_code=part_code,
                        context=context,
                        key="sold_quantity",
                        daily_goal_index=index,
                    )

            if not added_tickets:
                operator_daily_goal.tickets += number_of_seats
                added_tickets = True

        self._update_goal_percentage(
            item_goals=daily_goals.store_goals.item_goals,
            is_mem=True,
            store_goal=True,
        )

        return added_tickets

    @staticmethod
    def _check_part_code_context(
        part_code,      # type: str
        context,        # type: str
        part_codes,     # type: List
    ):
        # type: (...) -> bool

        for item in part_codes:
            context_match = contexts_match(
                item_context=str(context),
                part_code_context=str(item.context),
            )
            if part_code == item.id and (item.context in [None, ""] or context_match):
                return True

        return False

    def _update_sold_info(
        self,
        line,               # type: eTree.XML
        daily_goal,         # type: Any[Union[OperatorGoals, StoreGoals]]
        part_code,          # type: str
        context,            # type: str
        key,                # type: str
        daily_goal_index,   # type: int
    ):
        # type: (...) -> Any[Union[OperatorGoals, StoreGoals]]

        if self._check_part_code_context(
            part_code=part_code,
            context=context,
            part_codes=daily_goal.item_goals[daily_goal_index].part_codes,
        ):
            daily_goal.item_goals[daily_goal_index].quantity_sold += float(line.get("multipliedQty"))
            price = line.get("unitPrice") or line.get("addedUnitPrice") or 0.0

            if line.get("multipliedQty") is None:
                return daily_goal

            if key == "amount_sold":
                daily_goal.amount_sold += float(line.get("multipliedQty")) * float(price)
            elif key == "sold_quantity":
                daily_goal.sold_quantity += float(line.get("multipliedQty")) * float(price)
                self._update_sold_part_codes(
                    daily_goal_item=daily_goal.item_goals[daily_goal_index],
                    context=context,
                    part_code=part_code,
                    qty=float(line.get("multipliedQty")),
                )

        return daily_goal

    @staticmethod
    def _update_sold_part_codes(
        daily_goal_item,    # type: ItemGoals
        context,            # type: str
        part_code,          # type: str
        qty,                # type: float
    ):
        # type: (...) -> None

        part_codes_type = [daily_goal_item.part_codes, daily_goal_item.part_codes_set]
        for part_code_item in part_codes_type:
            for item_part_code in part_code_item:
                context_match = contexts_match(
                    item_context=str(context),
                    part_code_context=str(item_part_code.context),
                )
                if context_match and part_code == item_part_code.id:
                    if item_part_code.quantity_sold is None:
                        item_part_code.quantity_sold = 0
                    item_part_code.quantity_sold += qty

    def _populate_operators_sold_values(
        self,
        end_date,           # type: datetime
        initial_date,       # type: datetime
        store_goals,        # type: StoreGoals
        item_goals,         # type: List[ItemGoals]
        operators_goals,    # type: List[OperatorGoals]
    ):
        # type: (...) -> None

        users_tickets = self._get_tickets(business_period=initial_date)
        logger.info(users_tickets)

        for op in self._get_all_opened_users():
            operator_item_goals = []
            for item_goal in item_goals:
                operator_item_goals.append(ItemGoals(
                    name=item_goal.name,
                    part_codes=item_goal.part_codes,
                    part_codes_set=item_goal.part_codes_set,
                    quantity_goal=item_goal.quantity_goal,
                    quantity_operator_goal=item_goal.quantity_operator_goal,
                    quantity_sold=0,
                ))

            user_tickets = list({int(v) for k, v in users_tickets.items() if k.startswith(op)})
            user_tickets = 0 if len(user_tickets) == 0 else user_tickets[0]

            operator_goals = OperatorGoals(
                op,
                store_goals.operator_sale_goal if store_goals is not None else 0,
                0,
                store_goals.average_sale_goal if store_goals is not None else 0,
                operator_item_goals,
                user_tickets
            )

            operator_goals.amount_sold = self._get_current_sold_qty_items_from_database(
                items=operator_goals.item_goals,
                initial_date=initial_date,
                end_date=end_date,
                operator_id=op,
                store_goal=False,
            )
            self._update_goal_percentage(
                item_goals=operator_goals.item_goals,
                operator_id=op,
                store_goal=False,
            )
            operators_goals.append(operator_goals)

    def _get_tickets(
        self,
        business_period,    # type: datetime
    ):
        # type: (...) -> Dict

        date_format = "%Y-%m-%d"
        business_period = business_period.strftime(date_format)
        user_table_orders = self.order_repository.get_paid_ticket(business_period=business_period)
        user_orders = self.order_repository.get_paid_orders_ticket(business_period=business_period.replace("-", ""))
        tickets = {}
        for table_order in user_table_orders:
            user = table_order[0]
            quantity = int(table_order[1])
            tickets[user] = quantity

        for order in user_orders:
            if order[0] is None:
                continue

            user = order[0].split("user=")[1].split(",")[0]
            quantity = int(order[1])
            if user in tickets:
                tickets[user] += quantity
                continue
            tickets[user] = quantity

        return tickets

    def _populate_store_sold_values(
        self,
        end_date,       # type: datetime
        initial_date,   # type: datetime
        operator_id,    # type: str
        store_goals,    # type: StoreGoals
    ):
        # type: (...) -> None

        operator_id = "0" if operator_id in ["", "None"] else operator_id
        store_goals.sold_quantity = self._get_current_sold_qty_items_from_database(
            items=store_goals.item_goals,
            initial_date=initial_date,
            end_date=end_date,
            operator_id=operator_id,
            store_goal=True,
        )
        self._update_goal_percentage(
            item_goals=store_goals.item_goals,
            store_goal=True,
        )

    def _add_new_operator_to_mem(
        self,
        daily_goals,    # type: Goals
        operator_id,    # type: int
    ):
        # type: (...) -> None

        store_goals = self.today_goals.store_goals
        items_goals = self.today_goals.item_goals
        operator_item_goals = []
        for item_goal in items_goals:
            operator_item_goals.append(
                ItemGoals(
                    item_goal.name,
                    item_goal.part_codes,
                    item_goal.quantity_goal,
                    item_goal.quantity_operator_goal,
                    0
                )
            )
        operator_goals = OperatorGoals(
            operator_id,
            store_goals.operator_sale_goal,
            0,
            store_goals.average_sale_goal,
            operator_item_goals,
            0
        )
        daily_goals.operator_goals.append(operator_goals)

    @staticmethod
    def _get_all_opened_users():
        # type: () -> List
        users_list = []

        for user in eTree.XML(get_user_information()).findall(".//user"):
            users_list.append(user.get("UserId"))

        return users_list

    def _get_current_sold_qty_items_from_database(
        self,
        items,              # type: List
        initial_date,       # type: datetime
        end_date,           # type: datetime
        operator_id,        # type: str
        key="part_codes",   # type: Optional[str]
        store_goal=False,   # type: Optional[bool]
    ):
        # type: (...) -> Union[list, Union[list, float]]

        logger.info('[_get_current_sold_qty_items_from_database] Begin')

        total_sold = 0
        order_items = json.loads(
            self._get_order_items_from_database(
                initial_date=initial_date,
                end_date=end_date,
                operator_id=operator_id,
            )
        )

        is_part_code_set = key == 'part_codes_set'
        for item in items:
            total_sold_item, quantity_sold_item = self._get_sold_info(
                part_codes=item[key],
                order_items=order_items,
                store_goal=store_goal,
            )
            if not is_part_code_set:
                item.quantity_sold += quantity_sold_item
                total_sold += total_sold_item

        logger.info('[_get_current_sold_qty_items_from_database] End')

        if is_part_code_set:
            return items[0]

        return total_sold

    @staticmethod
    def _get_sold_info(
        part_codes,         # type: List
        order_items,        # type: List
        store_goal=False,   # type: Optional[bool]
    ):
        # type: (...) -> Tuple

        total_sold = 0
        quantity_sold = 0
        for order_item in order_items:
            for part_code in part_codes:
                if int(part_code.id) != order_item['pcode']:
                    continue
                context_match = contexts_match(
                    item_context=str(order_item['context']),
                    part_code_context=str(part_code.context),
                )
                if part_code.context is None or context_match:
                    quantity_sold += order_item['quantity']
                    unit_price = float(order_item["unit_price"] if order_item["unit_price"] != "None" else 0.0)
                    total_sold += order_item['quantity'] * unit_price
                    if store_goal:
                        if part_code.quantity_sold is None:
                            part_code.quantity_sold = 0
                        part_code.quantity_sold += order_item['quantity']

        return total_sold, quantity_sold

    def _get_order_items_from_database(
        self,
        initial_date,   # type: str
        end_date,       # type: str
        operator_id,    # type: str
    ):
        # type: (...) -> str

        paid_orders = self.order_repository.get_paid_orders(
            operator_id=operator_id,
            initial_date=initial_date,
            end_date=end_date,
        )
        order_items = self.order_repository.get_orders_items(paid_orders=paid_orders)
        order_items = filter(lambda x: x != "error", order_items)
        order_items = self.product_repository.get_order_items_names(order_items=order_items)
        order_items = sorted(order_items, key=lambda order_item: order_item.context)
        return json.dumps(map(lambda x: x.__dict__, order_items))

    def _update_goal_percentage(
        self,
        item_goals,         # type: ItemGoals
        operator_id='0',    # type: Optional[str]
        is_mem=False,       # type: Optional[bool]
        store_goal=False,   # type: Optional[bool]
    ):
        # type: (...) -> None

        for item in item_goals:
            if not item.part_codes_set:
                continue

            part_code_set_info = item

            if not is_mem:
                initial_date = datetime.now()
                end_date = datetime.now()
                part_code_set_info = self._get_current_sold_qty_items_from_database(
                    items=[item],
                    initial_date=initial_date,
                    end_date=end_date,
                    operator_id=operator_id,
                    key='part_codes_set',
                    store_goal=store_goal,
                )

            item.total_goal_percentage = round(item.quantity_sold * 100 / item.quantity_goal, 2)
            quantity_sold_set = 0
            for part_code_set in part_code_set_info.part_codes_set:
                if part_code_set.quantity_sold is not None:
                    quantity_sold_set = part_code_set.quantity_sold + quantity_sold_set
            if item.quantity_sold != 0:
                item.part_code_set_percentage = round(quantity_sold_set * 100 / item.quantity_sold, 2)
            else:
                item.part_code_set_percentage = 0



from abc import (
    ABCMeta,
    abstractmethod
)
from typing import (
    List,
    Tuple
)


class FiscalRepository(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_orders(
            self,
            end_date    # type: int
    ):
        # type: (...) -> List[Tuple[int, str]]

        raise NotImplementedError()

    @abstractmethod
    def fill_fiscal_backup_database(
            self,
            suffix_date,    # type: str
            orders          # type: List[int]
    ):
        # type: (...) -> None

        raise NotImplementedError()

    @abstractmethod
    def delete_fiscals(
            self,
            orders  # type: List[int]
    ):
        # type: (List[int]) -> None

        raise NotImplementedError()



from common import TipBuilderCommon
from pos_model import Order, SaleItem
from typing import Dict, Any

from _ContextKeys import ContextKeys
from _SatXmlPartBuilder import SatXmlPartBuilder


class DetTipBuilder(SatXmlPartBuilder):

    def __init__(self, det_builder, tip_config, tip_builder_common):
        # type: (SatXmlPartBuilder, dict, TipBuilderCommon) -> None
        self.det_builder = det_builder
        self.tip_config = tip_config
        self.tip_builder_common = tip_builder_common

    def build_xml(self, order, context):
        # type: (Order, Dict[unicode, Any]) -> unicode
        xml = ""
        tip_items = self.tip_builder_common.get_tip_items(order)
        for tip_item in tip_items:
            context[ContextKeys.current_sale_item] = tip_item
            xml += self.det_builder.build_xml(order, context)
            context[ContextKeys.item_no] += 1
        return xml
